"""
Module to take care of reading in inputs
wither from files or as direct objects
"""

import logging
import os

import numpy as np
import pandas as pd

# from ._utils import check_numeric_pamset
from ._utils import cut_and_check_pamset

LOGGER = logging.getLogger(__name__)

default_data_dir = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "default_data"
)


def read_components(filename):
    """
    Read in components to be considered

    Read the gas_pam files and rename a few headers to
    make it easier to use

    Parameters
    ----------
    filename : int
            path to gaspamfile

    Returns
    -------
    pandas.Dataframe
        Dataframe with gases, and various info on them
    """
    df_gas = pd.read_csv(filename, delim_whitespace=True, index_col=0)
    df_gas.rename(
        columns={"TAU1(YEARS)": "TAU1", "NATURAL_EMISSIONS": "NAT_EM"}, inplace=True
    )
    return df_gas


def read_natural_emissions(filename, component, startyear=1750, endyear=2500):
    """
    Read in single column natural emissions file

    A natural emissions file with data on a single column is read in
    to a pandas Dataframe. A an index of corresponding years is
    generated and added. Data in input is assumed to be yearly.

    Parameters
    ----------
    filename : str
               path to file with emissions
    component : str
                Name of component the emissions are for
    startyear: int
               Startyear for emissions
    endyear : int
              Endyear for emissions

    Returns
    -------
    pd.Dataframe
                Dataframe of natural emissions for component with years as index
    """
    df_natemis = pd.read_csv(filename, header=None, names=[component], index_col=False)
    df_natemis["year"] = np.arange(startyear, endyear + 1)
    df_natemis = df_natemis.set_index("year")
    return df_natemis


def read_inputfile(input_file, cut_years=False, year_start=1750, year_end=2100):
    """
    Read input from emissions or concentrations file

    Parameters
    ----------
    input_file : str
              Path to file to be read in
    cut_years : bool
             If unused years are to be cut, this option should
             be set to True, default is False
    year_start : int
             Start year for relevant input, default is 1750
    year_end : int
             End year for relevant input, default is 2100

    Returns
    -------
    pandas.Dataframe
        Dataframe with the intput from the file, possibly cut
        to relevant years
    """
    df_input = pd.read_csv(
        input_file, delim_whitespace=True, index_col=0, skiprows=[1, 2, 3]
    )
    if cut_years:
        min_year = df_input.index[0]
        max_year = df_input.index[0]
        cut_rows = [*range(min_year, year_start), *range(year_end + 1, max_year + 1)]
        df_input.drop(index=cut_rows, inplace=True)
    return df_input


def check_inputfiles(cfg):
    """
    Check whether input files are present or not

    Checking configuration dictionary to see whether
    it necessary files for concentrations or
    emissions run are present.
    If natural emissions files are not found in cfg,
    standard file location is used.

    Parameters
    ----------
    cfg : dict
       Configurations dictionary which should contain
       locations of necessary files.

    Returns
    -------
    dict
        cfg possible augmented with standard locations
        for natural emissions files

    Raises
    ------
    FileNotFoundError
         If files are not found
    """
    with_defaults = ["nat_ch4_file", "nat_n2o_file"]
    for key in cfg:
        if key[-4:] == "file":
            if not os.path.exists(cfg[key]):
                if key not in with_defaults:
                    raise FileNotFoundError(f"Input file {key} not found at {cfg[key]}")
                fname = f"natemis_{key.split('_')[1]}.txt"
                cfg[key] = os.path.join(os.getcwd(), "input_OTHER", "NATEMIS", fname)
    return cfg


def read_forc(forc_file):
    """
    Read in forcing from forc_file

    Read in forcing file to dataframe, couple of options
    depending on file formatting

    Parameters
    ----------
    forc_file : str
             Full path of forcing file to be read

    Returns
    -------
    ndarray
           Forcing data, or possibly a  pandas.Dataframe if
           data is organized in several components

    """
    components = False
    with open(forc_file, "r", encoding="utf8") as fread:
        first_line = fread.readline()
        if first_line[:4].lower() == "year":
            components = True
    if not components:
        df_forc = np.loadtxt(forc_file)
    else:
        # Decide on formatting for this
        df_forc = pd.read_csv(forc_file, index_col=0)
        if "total" not in df_forc.columns:
            if "FORC_NH" in df_forc.columns and "FORC_SH" in df_forc.columns:
                df_forc["total"] = df_forc[["FORC_NH", "FORC_SH"]].mean()
            else:
                df_forc["total"] = df_forc[list(df_forc)].sum(axis=1)
                df_forc["FORC_NH"] = df_forc["total"]
                df_forc["FORC_SH"] = df_forc["total"]
    return df_forc


def read_csv_no_index_col(filename):
    """
    Read input with pandas with no index column

    A method to wrap pandas.read_csv to have index_col = None
    so it can be called as a standard read in method for
    InputHandler.get_data

    Parameters
    ----------
    filename : str
              Path to file to be read in

    Returns
    -------
    pandas.Dataframe
        Dataframe with the intput from the file, possibly cut
        to relevant years
    """
    return pd.read_csv(filename, index_col=None)


class InputHandler:
    """
    Class to handle input sources of different kinds
    and taking care of how to get the correct input
    data
    """

    def __init__(self, cfg):
        """
        Initialise input handler

        Parameters
        ----------
        cfg : dict
           configurations such as input file locations
           Whether run is concentration run, if perturbations
           should be included etc.

        Raises
        ------
        FileNotFoundError
            If forcing file is not found when forcing run is chosen
        """
        self.read_methods = {
            "forc": read_forc,
            "gaspam": read_components,
            "nat_ch4": [read_natural_emissions, "CH4"],
            "nat_n2o": [read_natural_emissions, "N2O"],
            "concentrations": [
                read_inputfile,
                True,
                1750,
                2100,
            ],
            "emissions": self.read_emissions,
            "perturb_em": read_csv_no_index_col,
            "perturb_forc": pd.read_csv,
            "rf_luc": self.read_data_on_year_row,
            "rf_sun": self.read_data_on_year_row,
            "rf_volc_n": self.read_data_on_year_row,
            "rf_volc_s": self.read_data_on_year_row,
        }
        used = {
            "sunvolc": 0,
            "conc_run": False,
            "idtm": 24,
        }
        for key in self.read_methods:
            used[f"{key}_file"] = 0
            used[f"{key}_data"] = 0
        self.cfg = cut_and_check_pamset(
            {"nystart": 1750, "nyend": 2100, "emstart": 1850},
            cfg,
            used=used,
            cut_warnings=True,
        )
        self.read_methods["concentrations"] = [
            read_inputfile,
            True,
            self.cfg["nystart"],
            self.cfg["nyend"],
        ]
        self.set_sun_volc_luc_defaults()

        for nat_emis in ["nat_ch4", "nat_n2o"]:
            if not self.optional_pam(nat_emis) and self.optional_pam("gaspam"):
                self.cfg[f"{nat_emis}_data"] = self.get_flat_natural_emissions(
                    nat_emis.split("_")[1].upper()
                )

        check_inputfiles(self.cfg)

    def get_flat_natural_emissions(self, component):
        """
        Read flat natural emissions from gaspam_file
        if natural emissions are not provided as file or data

        Parameters
        ----------
        component : str
                    String to denote the component, should be
                    N2O or CH4

        Returns
        -------
        pd.Dataframe
                Dataframe of natural emissions for component with years as index
        """
        df_gas = self.get_data("gaspam")
        value = df_gas.at[component, "NAT_EM"]
        years = np.arange(self.cfg["nystart"], self.cfg["nyend"] + 1)
        return pd.DataFrame(data={component: np.ones(len(years)) * value}, index=years)

    def set_sun_volc_luc_defaults(self):
        """
        Set default values for landuse change, solar and
        volcanic data
        """
        if "rf_luc_file" not in self.cfg and "rf_luc_data" not in self.cfg:
            self.cfg["rf_luc_file"] = os.path.join(
                default_data_dir, "IPCC_LUCalbedo.txt"
            )

        if ("sunvolc", 1) in self.cfg.items():
            # Possibly change to allow for other files
            # And for SH to differ from NH
            if "rf_sun_file" not in self.cfg and "rf_sun_data" not in self.cfg:
                self.cfg["rf_sun_file"] = os.path.join(
                    default_data_dir, "solar_IPCC.txt"
                )
            volc_options = [
                "rf_volc_file",
                "rf_volc_data",
                "rf_volc_n_file",
                "rf_volc_s_file",
                "rf_volc_n_data",
                "rf_volc_s_data",
            ]
            intersection = [i for i in volc_options if i in self.cfg]
            if not intersection:
                self.cfg["rf_volc_n_file"] = os.path.join(
                    default_data_dir, "meanVOLCmnd_ipcc_NH.txt"
                )
                self.cfg["rf_volc_s_file"] = os.path.join(
                    default_data_dir, "meanVOLCmnd_ipcc_NH.txt"
                )
            elif "rf_volc_file" in self.cfg:
                self.cfg["rf_volc_n_file"] = self.cfg["rf_volc_file"]
                self.cfg["rf_volc_s_file"] = self.cfg["rf_volc_file"]
            elif "rf_volc_data" in self.cfg:
                self.cfg["rf_volc_n_data"] = self.cfg["rf_volc_data"]
                self.cfg["rf_volc_s_data"] = self.cfg["rf_volc_data"]
        else:
            indices = np.arange(self.cfg["nystart"], self.cfg["nyend"] + 1)
            self.cfg["rf_volc_n_data"] = pd.DataFrame(
                data=np.zeros((self.cfg["nyend"] - self.cfg["nystart"] + 1, 12)),
                index=indices,
                columns=range(12),
            )
            self.cfg["rf_volc_s_data"] = self.cfg["rf_volc_n_data"]
            self.cfg["rf_sun_data"] = pd.DataFrame(
                data=np.zeros(self.cfg["nyend"] - self.cfg["nystart"] + 1),
                index=indices,
            )

    def get_rf_type(self):
        """
        Check if this is rf_run

        Checking if an input key with rf_run info
        is contained in the configuration set, i.e.
        the dataset is an rf_run

        Returns
        -------
        bool
            Whether this is rf_run or not
        """
        rf_type = False
        if "forc_file" in self.cfg or "forc_data" in self.cfg:
            rf_type = True
        return rf_type

    def get_data(self, what):
        """
        Get data from file or sent directly as data

        Method that according to what is looked for
        searches the configuration set self.cfg for
        a parameter that has the same name with _file
        at the end, and if so, looks up the correct
        read in method for that file, or if for
        a parameter with the name with _data at the end,
        in which case it returns that dataset.
        If there is no reading method when name_file
        is present, or if there is no name_file or
        name_data in configurations, then a KeyError
        is raised.

        Parameters
        ----------
        what : str
             what is being looked for, such as emissions,
             concentrations, gaspam etc.

        Returns
        -------
               Dataframe or numpyarray

        Raises
        ------
        KeyError
               If data for what is named is not in configurations
               or if there is no read_method for a file to be read
        """
        if f"{what}_file" in self.cfg:
            if what not in self.read_methods:
                raise KeyError(f"No reading method available for {what}_file")
            if isinstance(self.read_methods[f"{what}"], list):
                return self.read_methods[f"{what}"][0](
                    self.cfg[f"{what}_file"], *self.read_methods[f"{what}"][1:]
                )
            return self.read_methods[f"{what}"](self.cfg[f"{what}_file"])
        if f"{what}_data" in self.cfg:
            # Sanity check of input data?
            return self.cfg[f"{what}_data"]
        raise KeyError(f"No user or default data for {what}")

    def read_emissions(self, filename):
        """
        Read in emission from file

        Method to read in emissions data file by
        calling read_inputfile method and renaming
        CO2 columns

        Parameters
        ----------
        filename : str
             path to emission file to be read in

        Returns
        -------
               pandas.Dataframe
        """
        emis = read_inputfile(
            filename,
            cut_years=True,
            year_start=self.cfg["nystart"],
            year_end=self.cfg["nyend"],
        )
        emis.rename(columns={"CO2": "CO2_FF", "CO2.1": "CO2_AFOLU"}, inplace=True)
        return emis

    def conc_run(self):
        """
        Check if configurations includes conc_run option

        Checking if configurations includes conc_run option
        If included, its value is returned, otherwise False
        is returned

        Returns
        -------
        bool
            Whether this is rf_run or not
        """
        if "conc_run" in self.cfg:
            return self.cfg["conc_run"]
        return False

    def optional_pam(self, which):
        """
        Check if data that can be sent optionally is included

        Checking if an input key for an optionally input data
        is found in the configurations either as file or data
        This could be perturb_em, pertub_forc or forc

        Parameters
        ----------
        which : str
               Type of data to be checked for. This could be
               perturb_em, perturb_forc or forc

        Returns
        -------
        bool
            Whether this is rf_run or not
        """
        return bool(f"{which}_file" in self.cfg or f"{which}_data" in self.cfg)

    def read_data_on_year_row(self, volc_datafile):
        """
        Read in data from file with no headers


        Read in data from file with no headers where
        each year is a row. Typically this is the format for
        volcano and solar data. The years are taken to be
        the years from the defined startyear and endyear

        Parameters
        ----------
        volc_datafile : str
                     Path of file to be read

        Returns
        -------
        pandas.Dataframe
                        Dataframe containing the data with the years as
                        indices
        """
        indices = np.arange(self.cfg["nystart"], self.cfg["nyend"] + 1)
        nrows = len(indices)
        if self.cfg["nystart"] > 1750:
            skiprows = self.cfg["nystart"] - 1750
            df_data = pd.read_csv(
                volc_datafile,
                header=None,
                skiprows=skiprows,
                nrows=nrows,
                delim_whitespace=True,
            )
        else:
            df_data = pd.read_csv(
                volc_datafile, header=None, nrows=nrows, delim_whitespace=True
            )

        df_data.set_axis(labels=indices)
        return df_data



"""
Module to handle emission, concentrations and converstion to forcing
"""

import logging
import os

import numpy as np
import pandas as pd

# from ._utils import check_numeric_pamset
from ._utils import cut_and_check_pamset
from .make_plots import plot_output2
from .perturbations import (
    ForcingPerturbation,
    calculate_hemispheric_forcing,
    perturb_emissions,
)
from .pub_utils import make_cl_and_br_dictionaries

LOGGER = logging.getLogger(__name__)


def _rs_function(it, idtm=24):
    """
    Calculate pulse response function for mixed layer

    Calculate pulse response function for mixed layer
    time is the year index*idtm + i, i.e. the month number

    Parameters
    ----------
    it : int
      is the time index, there are idtm time points per yer
    idtm : int
        Number of time points per year, default is 24

    Returns
    -------
    float
         The pulse_response function for this time
    """
    time = it / idtm
    if time < 2.0:
        pulse_response = (
            0.12935
            + 0.21898 * np.exp(-time / 0.034569)
            + 0.17003 * np.exp(-time / 0.26936)
            + 0.24071 * np.exp(-time / 0.96083)
            + 0.24093 * np.exp(-time / 4.9792)
        )
    else:
        pulse_response = (
            0.022936
            + 0.24278 * np.exp(-time / 1.2679)
            + 0.13963 * np.exp(-time / 5.2528)
            + 0.089318 * np.exp(-time / 18.601)
            + 0.03782 * np.exp(-time / 68.736)
            + 0.035549 * np.exp(-time / 232.3)
        )
    return pulse_response


def _rb_function(it, idtm=24):
    """
    Calculate biotic decay function

    Calculate biotic decay function
    time is the year index*idtm + i, i.e. the month number

    Parameters
    ----------
    it : int
      is the time index, there are idtm time points per yer
    idtm : int
        Number of time points per year, default is 24

    Returns
    -------
    float
        The biotic decay function value for this time
    """
    time = it / idtm
    biotic_decay = (
        0.70211 * np.exp(-0.35 * time)
        + 13.4141e-3 * np.exp(-time / 20.0)
        - 0.71846 * np.exp(-55 * time / 120.0)
        + 2.9323e-3 * np.exp(-time / 100.0)
    )
    return biotic_decay


def check_pamset(pamset):
    """
    Check that parameterset has necessary values for run

    Check that parameterset has necessary values for run
    Otherwise set to default values which are defined here

    Parameters
    ----------
    pamset : dict
          Dictionary of parameters to define the physics
          of the run. Values that begin with q are concetration
          or emissions to forcing factors, beta_f is the
          carbon cycle fertilisation factor, and ref_yr is
          the reference year for calculations

    Returns
    -------
    dict
        Updated pamset with default values used where necessary
    """
    required = {
        "qbmb": 0.0,
        "qo3": 0.5,
        "qdirso2": -0.36,
        "qindso2": -0.97,
        "qbc": 0.16,
        "qoc": -0.08,
        "qh2o_ch4": 0.091915,
        "ref_yr": 2010,
        "beta_f": 0.287,
    }

    # pamset = check_numeric_pamset(required, pamset, )
    if "lifetime_mode" not in pamset:
        pamset["lifetime_mode"] = "TAR"

    used = {
        "lifetime_mode": "TAR",
        "just_one": "CO2",
        "idtm": 24,
        "nystart": 1750,
        "nyend": 2100,
        "emstart": 1850,
    }
    pamset = cut_and_check_pamset(required, pamset, used=used, cut_warnings=True)
    return pamset


def check_pamset_consistency(pamset_old, pamset_new):
    """
    Make sure new pamset is consistent with existing instance

    This method is meant to ensure consistency when rerunning
    with the same class instance. Changing to old values
    if they are different for a few values that can't change,
    or setting to old values if none are given

    Parameters
    ----------
    pamset_old : dict
              Original parameterset for instance
    pamset_new : dict
              New pamset for new run
    Returns
    -------
    dict
        The new pamset, augmented to fit the old one
    """
    pams_cant_change = ["idtm", "nystart", "nyend", "emstart", "conc_run"]
    for pam in pams_cant_change:
        if pam in pamset_new and pamset_new[pam] != pamset_old[pam]:
            LOGGER.warning(  # pylint: disable=logging-fstring-interpolation
                f"{pam} can not be changed for same instance of ConcentrationsEmisssionsHandler. Resetting with old value {pamset_old[pam]}. If you want to run with a different value, please create a separate instance",
            )
            pamset_new[pam] = pamset_old[pam]
    for pam, value in pamset_old.items():
        if pam not in pamset_new:
            pamset_new[pam] = value
    return pamset_new


class ConcentrationsEmissionsHandler:
    """
    Class to handle concentrations
    and emissions input for ciceroscm

    Attributes
    ----------
    df_gas : pd.Dataframe
             Dataframe containing names of the various components
             units, possible natural emissions and emissions to
             concentrations and concentrations to forcing conversion
             factors
    conc : dict
           Dictionary with component keys and arrays with concentrations
           for each year as values. These values are calculated
           when emi2conc method is called. Before emissions start
           or if a concentration run is used, they will be read
           directly from conc_in
    nat_emis_ch4 : pd.Dataframe
                   Natural emissions for CH4 per year
    nat_emis_n2o : pd.Dataframe
                   Natural emissions for N2O per year
    pamset : dict
             Dictionary of parameters
    years : np.ndarray
            Array with all the years for the handler
    conc_in : pd.Dataframe
              Input concentrations read from file. These are used
              before emissions start, or throughout the run if
              a concentration run is used
    emis : pd.Dataframe
           Emissions dataframe read from input file
    r_functions : np.ndarray
                  2D array with precalculated values of pulse
                  response and biotic decay functions for each
                  of the idtm values of each year


    """

    # pylint: disable=too-many-instance-attributes
    # Need a few more here or else consider breaking
    # up emissions in it's own class or
    # CO2-handling in its own class

    def __init__(
        self,
        input_handler,
        pamset,
    ):
        """
        Intialising concentrations emissions handler

        Starting by reading in gasparameter file, making
        empty dicts for concentrations and forcing,
        reading in natural emissions, checking the pamset,
        and making an array of years to look at.
        Then reading in concentrations and emissions,
        perturbations if relevant.
        Then initialising pulse and biotic functions,
        finally intialising empty arrays and setting
        CO2 start values.

        Parameters
        ----------
        input_handler : ciceroscm.InputHandler
                       input_handler that takes care of
                       configurations and reading in of
                       data from user
        pamset : dict
           list of physical parameters to define the run
        """
        self.df_gas = input_handler.get_data("gaspam")
        self.conc = {}
        self.forc = {}
        self.nat_emis_ch4 = input_handler.get_data("nat_ch4")
        self.nat_emis_n2o = input_handler.get_data("nat_n2o")
        self.pamset = cut_and_check_pamset(
            {"idtm": 24, "nystart": 1750, "nyend": 2100, "emstart": 1850},
            pamset,
            used={"rs_function": _rs_function, "rb_function": _rb_function},
            cut_warnings=True,
        )
        self.years = np.arange(self.pamset["nystart"], self.pamset["nyend"] + 1)
        self.conc_in = input_handler.get_data("concentrations")
        self.emis = input_handler.get_data("emissions")
        self.pamset["conc_run"] = input_handler.conc_run()
        if input_handler.optional_pam("perturb_em"):
            perturb_emissions(input_handler, self.emis)
        if input_handler.optional_pam("perturb_forc"):
            self.pamset["forc_pert"] = ForcingPerturbation(input_handler, self.years[0])
        self.precalc_r_functions()
        self.pamset["cl_dict"], self.pamset["br_dict"] = make_cl_and_br_dictionaries(
            self.df_gas.index
        )
        # not really needed, but I guess the linter will complain...
        self.reset_with_new_pams(pamset, preexisting=False)

    def reset_with_new_pams(self, pamset, preexisting=True):
        """
        Reset to run again with same emissions etc.

        Resetting arrays and co2 start values also
        making sure parameterset conforms with previously
        existing set if new run is being run

        Parameters
        ----------
        pamset : dict
              Dictionary of physical parameters
        preexisting : bool
                   Defining whether this is the first use of the class
                   or a new run with the same instance, that has to
                   conform to old values
        """
        if preexisting:
            new_pamset = check_pamset(pamset)
            self.pamset = check_pamset_consistency(self.pamset, new_pamset)
        years_tot = len(self.years)
        self.conc = {}
        self.forc = {}
        for tracer in self.df_gas.index:
            if tracer != "CO2.1":
                self.conc[tracer] = {}
                self.forc[tracer] = np.zeros(years_tot)
        self.forc["Total_forcing"] = np.zeros(years_tot)
        self.co2_hold = {
            "yCO2": 0.0,
            "xCO2": 278.0,
            "sCO2": np.zeros(self.pamset["idtm"] * years_tot),
            "emCO2_prev": 0.0,
            "dfnpp": np.zeros(self.pamset["idtm"] * years_tot),
            "ss1": 0.0,
            "sums": 0.0,
        }

    def precalc_r_functions(self):
        """
        Precalculate decay functions either
        sent in pamset or from default

        If functions are sent with keywords rs_function
        or rb_function in the pamset, these must take
        time and number of steps per year as input

        Parameters
        ----------
        pamset
        """
        years_tot = len(self.years)
        self.r_functions = np.empty(
            (2, self.pamset["idtm"] * years_tot)
        )  # if speedup, get this to reflect number of years
        if "rs_function" not in self.pamset:
            self.pamset["rs_function"] = _rs_function
        if "rb_function" not in self.pamset:
            self.pamset["rb_function"] = _rb_function
        self.r_functions[0, :] = [
            self.pamset["rs_function"](it, self.pamset["idtm"])
            for it in range(self.pamset["idtm"] * years_tot)
        ]
        self.r_functions[1, :] = [
            self.pamset["rb_function"](it, self.pamset["idtm"])
            for it in range(self.pamset["idtm"] * years_tot)
        ]

    def calculate_strat_quantities(self, yr):
        """
        Calculate sumcl and sumbr in stratosphere

        Calculating sum of stratospheric effects of
        chlorinated gase and sum of stratospheric effects
        of halon quantities.

        Parameters
        ----------
        yr : int
          Year for which to calculate

        Returns
        -------
        list
            Containing the sumcl, value for chlorinated gases
            and the sumbr value for halon gases.
        """
        yr0 = int(self.years[0])
        if yr <= yr0:
            return 0.0, 0.0
        sumcl = 0
        sumbr = 0
        for comp, mult in self.pamset["cl_dict"].items():
            num = self.conc[comp][yr] - self.conc[comp][yr0]
            # sumcl = sumcl + (mult * (self.conc[comp][yr] - self.conc[comp][yr0])) ** 1.7
            sumcl = sumcl + np.sign(num) * np.power(mult * abs(num), 1.7)
        for comp, mult in self.pamset["br_dict"].items():
            sumbr = sumbr + mult * (self.conc[comp][yr] - self.conc[comp][yr0])
        return sumcl, sumbr

    def calculate_forc_three_main(self, yr):  # pylint: disable=too-many-locals
        """
        Calculate forcings for components CO2, N2O and CH4

        Method to calculate the forcing from the main three componenents
        as these three are a bit intertwined. Using a coefficient matrix

        Parameters
        ----------
        yr : int
          Year for which to calculate

        Returns
        -------
        list
            Containing the total forcing from the three, and the
            combinded forcing value on each hemisphere
            tot_forc, forc_nh, forc_sh
        """
        # Setting some constants. What are they? Should they be parametrisable?
        # (Etminan I guess...)

        # Array with coefficients
        co2_n2o_ch4_coeff = np.array(
            [
                [-2.4e-7, 7.2e-4, -2.1e-4],
                [-8.0e-6, 4.2e-6, -4.9e-6],
                [-1.3e-6, -8.2e-6, None],
            ]
        )
        # a1 = -2.4e-7
        # b1 = 7.2e-4
        # c1 = -2.1e-4
        # a2 = -8.0e-6
        # b2 = 4.2e-6
        # c2 = -4.9e-6
        # a3 = -1.3e-6
        # b3 = -8.2e-6
        yr_0 = self.years[0]
        c0_n2o = self.conc["N2O"][yr_0]
        c_n2o = self.conc["N2O"][yr]
        c0_co2 = self.conc["CO2"][yr_0]
        c_co2 = self.conc["CO2"][yr]
        c0_ch4 = self.conc["CH4"][yr_0]
        c_ch4 = self.conc["CH4"][yr]
        q_co2 = (
            co2_n2o_ch4_coeff[0, 0] * (c_co2 - c0_co2) ** 2
            + co2_n2o_ch4_coeff[0, 1] * np.abs(c_co2 - c0_co2)
            + co2_n2o_ch4_coeff[0, 2] * 0.5 * (c0_n2o + c_n2o)
            + 5.36
        ) * np.log(
            c_co2 / c0_co2
        )  # Etminan et al 2016 RF

        # Note there were a few unused calculations in the original
        # for n2o and ch4 including som fmn values
        q_n2o = (
            co2_n2o_ch4_coeff[1, 0] * 0.5 * (c_co2 + c0_co2)
            + co2_n2o_ch4_coeff[1, 1] * 0.5 * (c_n2o + c0_n2o)
            + co2_n2o_ch4_coeff[1, 2] * 0.5 * (c_ch4 + c0_ch4)
            + 0.117
        ) * (np.sqrt(c_n2o) - np.sqrt(c0_n2o))
        q_ch4 = (
            co2_n2o_ch4_coeff[2, 0] * 0.5 * (c_ch4 + c0_ch4)
            + co2_n2o_ch4_coeff[2, 1] * 0.5 * (c_n2o + c0_n2o)
            + 0.043
        ) * (np.sqrt(c_ch4) - np.sqrt(c0_ch4))
        # Feedback factor: Smith et al 2018
        q_co2 = self.df_gas["SARF_TO_ERF"]["CO2"] * q_co2  # + FORC_PERT(yr_ix,trc_ix))
        q_n2o = self.df_gas["SARF_TO_ERF"]["N2O"] * q_n2o  # + FORC_PERT(yr_ix,trc_ix))
        q_ch4 = self.df_gas["SARF_TO_ERF"]["CH4"] * q_ch4  # + FORC_PERT(yr_ix,trc_ix))

        self.forc["CO2"][yr - yr_0] = q_co2
        self.forc["CH4"][yr - yr_0] = q_ch4
        self.forc["N2O"][yr - yr_0] = q_n2o
        tot_forc = q_co2 + q_n2o + q_ch4
        forc_nh, forc_sh = calculate_hemispheric_forcing(
            "three_main", tot_forc, 0.0, 0.0
        )

        return tot_forc, forc_nh, forc_sh

    def tropospheric_ozone_forcing(self, yr):
        """
        Calculate tropospheric ozone forcing

        Using the emissions of various other gases
        Before emissions start, the fossil fuel CO2
        change from the start is used
        After emissions start, concentrations of
        methane are combined with emissions of NOx, CO and NMVOC
        to calculate a total troposperic ozone forcing

        Parameters
        ----------
        yr : int
          Year for which to calculate

        Returns
        -------
        float
             tropospheric ozone forcing
        """
        emstart = self.pamset["emstart"]
        yr_0 = self.years[0]
        yr_emstart = emstart - yr_0
        yr_ix = yr - yr_0
        tracer = "TROP_O3"
        if yr_ix < yr_emstart:
            # Uses change in CO2_FF emissions
            if self.emis["CO2_FF"][self.pamset["ref_yr"]] != self.emis["CO2_FF"][yr_0]:
                q = (
                    (self.emis["CO2_FF"][yr] - self.emis["CO2_FF"][yr_0])
                    / (
                        self.emis["CO2_FF"][self.pamset["ref_yr"]]
                        - self.emis["CO2_FF"][yr_0]
                    )
                    * self.pamset["qo3"]
                )
            else:
                q = (
                    (self.emis["CO2_FF"][yr] - self.emis["CO2_FF"][yr_0])
                ) * self.pamset["qo3"]

        else:
            # ALOG(1700.0))  !Concentration in 2010 &
            self.conc[tracer][yr] = (
                30.0
                + 6.7
                * (
                    np.log(self.conc["CH4"][yr])
                    - np.log(self.conc_in["CH4"][self.pamset["ref_yr"]])
                )
                + 0.17
                * (self.emis["NOx"][yr] - self.emis["NOx"][self.pamset["ref_yr"]])
                + 0.0014
                * (self.emis["CO"][yr] - self.emis["CO"][self.pamset["ref_yr"]])
                + 0.0042
                * (self.emis["NMVOC"][yr] - self.emis["NMVOC"][self.pamset["ref_yr"]])
            )
            # RBS101115
            # IF (yr_ix.LT.yr_2010) THEN ! Proportional to TROP_O3 build-up
            # Rewritten a bit, place to check for differences...
            forc_pre_emstart = self.forc[tracer][yr_emstart - 1]
            value = self.conc[tracer][yr]
            value_0 = self.conc[tracer][self.pamset["emstart"]]
            q = forc_pre_emstart + (value - value_0) / (30.0 - value_0) * (
                self.pamset["qo3"] - forc_pre_emstart
            )
        return q

    def conc2forc(self, yr, rf_luc, rf_sun):  # pylint: disable=too-many-branches
        """
        Calculate forcing from concentrations

        Looping through all species and getting the
        forcing from their concentrations. First calling
        a method to calculate intial forcing for three main
        species, CO2, N2O and CH4. This will also initialise
        total forcing and hemispherically split forcing.
        Then looping over species
        calculated dirctly from emissions (SO2, SO4_IND, OC and BC)
        Then looping over other gases with alpha factor,
        then ozone in troposphere and stratosphere, and stratospheric water
        vapour, finally other species, and adding perturbations if
        present. For every species add to total forcing and hemispherically
        split forcings, finally add solar forcing.

        Parameters
        ----------
        yr : int
          Year fro which to calculate
        rf_luc : float
              Land use change forcing
        rf_sun : float
              Solar forcing

        Returns
        -------
        list
            Consisting of total forcing and hemispherically split
            forcing for the year.
            In this way: tot_forc, forc_nh, forc_sh
        """
        ref_emission_species = {
            "SO2": ["SO2", self.pamset["qdirso2"]],
            "SO4_IND": ["SO2", self.pamset["qindso2"]],
            "OC": ["OC", self.pamset["qoc"]],
            "BC": ["BC", self.pamset["qbc"]],
            "BMB_AEROS": ["BMB_AEROS_OC", self.pamset["qbmb"]],
        }
        # Intialising with the combined values from CO2, N2O and CH4
        tot_forc, forc_nh, forc_sh = self.calculate_forc_three_main(yr)
        yr_0 = self.years[0]
        # Finish per tracer calculations, add per tracer to printable df and sum the total
        for tracer in self.df_gas.index:
            if tracer in ["CO2", "N2O", "CH4"]:
                continue
            q = 0
            if tracer == "LANDUSE":
                q = rf_luc
            elif tracer in ref_emission_species:
                # Natural emissions
                # (after IPCC TPII on simple climate models, 1997)
                # enat = 42.0 Not used, why is this here?
                # Emission in reference year
                # SO2, SO4_IND, BC and OC are treated exactly the same
                # Only with emission to concentration factors differing
                # These are held in dictionary
                erefyr = (
                    self.emis[ref_emission_species[tracer][0]][self.pamset["ref_yr"]]
                    - self.emis[ref_emission_species[tracer][0]][yr_0]
                )
                if erefyr != 0.0:  # pylint: disable=compare-to-zero
                    frac_em = (
                        self.emis[ref_emission_species[tracer][0]][yr]
                        - self.emis[ref_emission_species[tracer][0]][yr_0]
                    ) / erefyr
                    q = ref_emission_species[tracer][1] * frac_em

            elif (
                tracer in self.df_gas.index
                and self.df_gas["ALPHA"][tracer] != 0  # pylint: disable=compare-to-zero
            ):
                q = (
                    (self.conc[tracer][yr] - self.conc[tracer][yr_0])
                    * self.df_gas["ALPHA"][tracer]
                    * self.df_gas["SARF_TO_ERF"][tracer]
                )  # +forc_pert
            elif tracer == "TROP_O3":
                q = (
                    self.tropospheric_ozone_forcing(yr)
                    * self.df_gas["SARF_TO_ERF"][tracer]
                )
            elif tracer == "STRAT_O3":
                sumcl, sumbr = self.calculate_strat_quantities(yr - 3)
                # Updated according to IPCC AR4.
                # Multiply by factor 0.287737 (=-0.05/-0.17377, AR4/SCM)
                q = (
                    -(
                        # self.pamset["qo3"]
                        0.05
                        / 0.17377
                        * (0.000552 * (sumcl) + 3.048 * sumbr)
                    )
                    / 1000.0
                    * self.df_gas["SARF_TO_ERF"][tracer]
                )
            elif tracer == "STRAT_H2O":
                q = (
                    self.pamset["qh2o_ch4"] * self.forc["CH4"][yr - yr_0]
                ) * self.df_gas["SARF_TO_ERF"][
                    tracer
                ]  # + FORC_PERT(yr_ix,trc_ix)
            elif tracer == "OTHER":
                # Possible with forcing perturbations for other
                # components such as contrails, cirrus etc...
                pass
            self.forc[tracer][yr - yr_0] = q  # + FORC_PERT(yr_ix,trc_ix)
            # Calculating hemispheric forcing:
            forc_nh, forc_sh = calculate_hemispheric_forcing(
                tracer, q, forc_nh, forc_sh
            )
            tot_forc = tot_forc + q
            # print("Forcer: %s, tot_forc: %f, FN: %f, FS: %f, q: %f"%(tracer, tot_forc, forc_nh, forc_sh, q)

        # Adding forcing perturbations if they exist:
        if "forc_pert" in self.pamset:
            if self.pamset["forc_pert"].check_if_year_in_pert(yr):
                tot_forc, forc_nh, forc_sh, self.forc = self.pamset[
                    "forc_pert"
                ].add_forcing_pert(tot_forc, forc_nh, forc_sh, self.forc, yr)

        # Adding solar forcing
        # tot_forc = tot_forc + rf_sun
        if "just_one" in self.pamset:
            tot_forc = self.forc[self.pamset["just_one"]][yr - yr_0]
            forc_nh, forc_sh = calculate_hemispheric_forcing(
                self.pamset["just_one"], tot_forc, 0, 0
            )
        self.forc["Total_forcing"][yr - yr_0] = tot_forc
        forc_nh = forc_nh + rf_sun
        forc_sh = forc_sh + rf_sun

        return tot_forc, forc_nh, forc_sh

    def emi2conc(self, yr):
        """
        Calculate concentrations from emissions

        If conc_run or prior to emissions start, concentrations
        are simply read in from existing data frame, after that
        all tracers are looped over to calculate concentrations
        from emissions. For CO2 this is doen in a separate method.
        For CH4 and N2O, natural emissions are added.

        Parameters
        ----------
        yr : int
          Year for which to calculate
        """
        # Do per tracer emissions to concentrations, update concentrations df
        # NBNB! Remember to move calculation of  Trop_O3 concentration
        # here from conc2forc.
        # First calculate O3 concentrations (AS IN TAR p269, table 4.11 B)
        # CONC(yr_ix,trc_ix) =  30.0 + 6.7 * &
        # (ALOG(CONC(yr_ix,trcID("CH4")))-ALOG(1832.0)) &   !ALOG(1700.0))  !Concentration in 2010 &
        #  + 0.17 * (EMISSIONS(yr_ix,trcID("NOx"))-EM2010(trcID("NOx"))) &
        #  + 0.0014 * (EMISSIONS(yr_ix,trcID("CO"))-EM2010(trcID("CO"))) &
        #  + 0.0042 *(EMISSIONS(yr_ix,trcID("NMVOC"))-EM2010(trcID("NMVOC")))
        if self.pamset["conc_run"]:
            self.fill_one_row_conc(yr)
            return
        # Before emissions start
        if yr < self.pamset["emstart"]:
            self.co2em2conc(yr)
            self.fill_one_row_conc(yr, avoid=["CO2"])
            return
        self.add_row_of_zeros_conc(yr)

        for tracer in self.df_gas.index:
            # something something postscenario...?
            if self.df_gas["CONC_UNIT"][tracer] == "-":
                # Forcing calculated from emissions
                continue
            if tracer == "CO2":
                self.co2em2conc(yr)
                continue
            if yr < self.pamset["emstart"]:
                self.fill_one_row_conc(yr)

            if yr > self.years[0]:
                conc_local = self.conc[tracer][yr - 1]
            else:
                conc_local = self.conc_in[tracer][yr]

            q = 1.0 / self.df_gas["TAU1"][tracer]

            if tracer == "CH4":
                self.df_gas.at[tracer, "NAT_EM"] = self.nat_emis_ch4["CH4"][yr]
                q = self.methane_lifetime(q, conc_local, yr)
            if tracer == "N2O":
                self.df_gas.at[tracer, "NAT_EM"] = self.nat_emis_n2o["N2O"][yr]

            # natural emissions, from gasspamfile
            emis = self.emis[tracer][yr] + self.df_gas["NAT_EM"][tracer]

            point_conc = emis / self.df_gas["BETA"][tracer]
            # Rewrote this quite a bit from an original loop,
            # but I think it is mathematically equivalent
            self.conc[tracer][yr] = point_conc / q + (
                conc_local - point_conc / q
            ) * np.exp(-q)

    def methane_lifetime(self, q, conc_local, yr):
        """
        Calculate methane concentrations from emissions

        Calculate methane concentrations from emissions
        using different liftetime modes to calculate

        Parameters
        ----------
        q : float
         Forcing without feedbacks
        conc_local : float
                  Concentration of methane in previous timestep
        yr : int
          Year for which to calculate
        Returns
        -------
        float
             Concentrations adjusted for lifetime / feedback
        """
        ch4_wigley_exp = -0.238
        if self.pamset["lifetime_mode"] == "TAR":
            # 1751 is reference conc in 2000
            dln_oh = (
                -0.32 * (np.log(conc_local) - np.log(1751.0))
                + 0.0042 * (self.emis["NOx"][yr] - self.emis["NOx"][2000])
                - 0.000105 * (self.emis["CO"][yr] - self.emis["CO"][2000])
                - 0.000315 * (self.emis["NMVOC"][yr] - self.emis["NMVOC"][2000])
            )
            q = q * (dln_oh + 1)
        elif self.pamset["lifetime_mode"] == "CONSTANT_12":
            q = 1.0 / 12.0
        elif self.pamset["lifetime_mode"] == "WIGLEY":
            q = q * (((conc_local / 1700.0)) ** (ch4_wigley_exp))

        q = q + 1.0 / self.df_gas["TAU2"]["CH4"] + 1.0 / self.df_gas["TAU3"]["CH4"]

        return q

    def co2em2conc(self, yr):  # pylint: disable=too-many-locals
        """
        Calculate co2 concentrations from emissions

        Method to calculate co2 concentrations from emissions
        Implementing a rudimentary carbon cycle which loops over
        idtm (usually 24) timesteps a year

        Parameters
        ----------
        yr : int
          Year for which to calculate
        """
        # Area of the ocean (m^2)
        ocean_area = 3.62e14

        # Gas exchange coefficient (air <--> ocean) (yr^-1*m^-2)
        coeff = 1.0 / (ocean_area * 9.06)

        # TIMESTEP (YR)
        dt = 1.0 / self.pamset["idtm"]

        # Conversion factor ppm/kg --> umol*/m3
        conv_factor = 1.722e17

        # USING MIXED LAYER DEPTH = 75 metres
        mixed_layer_depth = 75.0

        cc1 = dt * ocean_area * coeff / (1 + dt * ocean_area * coeff / 2.0)
        yr_ix = yr - self.years[0]
        # Monthloop:
        em_co2_common = (
            self.emis["CO2_FF"][yr]
            + self.emis["CO2_AFOLU"][yr]
            + self.df_gas["NAT_EM"]["CO2"]
        )
        for i in range(self.pamset["idtm"]):
            it = yr_ix * self.pamset["idtm"] + i
            sumf = 0.0

            # Net emissions, including biogenic fertilization effects
            if it > 0:
                self.co2_hold["dfnpp"][it] = (
                    60 * self.pamset["beta_f"] * np.log(self.co2_hold["xCO2"] / 278.0)
                )
            if it > 0:
                sumf = float(
                    np.dot(
                        self.co2_hold["dfnpp"][1:it],
                        np.flip(self.r_functions[1, : it - 1]),
                    )
                )

            ffer = self.co2_hold["dfnpp"][it] - dt * sumf
            em_co2 = (em_co2_common - ffer) / 2.123

            if it == 0:  # pylint: disable=compare-to-zero
                self.co2_hold["ss1"] = 0.5 * em_co2 / (ocean_area * coeff)
                ss2 = self.co2_hold["ss1"]
                self.co2_hold["sums"] = 0.0
            else:
                ss2 = 0.5 * em_co2 / (ocean_area * coeff) - self.co2_hold["yCO2"] / (
                    dt * ocean_area * coeff
                )
                self.co2_hold["sums"] = (
                    self.co2_hold["sums"]
                    + self.co2_hold["emCO2_prev"] / (ocean_area * coeff)
                    - self.co2_hold["sCO2"][it - 1]
                )
            self.co2_hold["sCO2"][it] = cc1 * (
                self.co2_hold["sums"] + self.co2_hold["ss1"] + ss2
            )
            self.co2_hold["emCO2_prev"] = em_co2
            if it > 0:
                sumz = np.dot(
                    self.co2_hold["sCO2"][: it - 1], np.flip(self.r_functions[0, 1:it])
                )
            else:
                sumz = 0.0

            z_co2 = (
                conv_factor
                * coeff
                * dt
                / mixed_layer_depth
                * (sumz + 0.5 * self.co2_hold["sCO2"][it])
            )
            self.co2_hold["yCO2"] = (
                1.3021 * z_co2
                + 3.7929e-3 * (z_co2**2)
                + 9.1193e-6 * (z_co2**3)
                + 1.488e-8 * (z_co2**4)
                + 1.2425e-10 * (z_co2**5)
            )
            self.co2_hold["xCO2"] = (
                self.co2_hold["sCO2"][it] + self.co2_hold["yCO2"] + 278.0
            )
            # print("it: %d, emCO2: %e, sCO2: %e, zCO2: %e, yCO2: %e, xCO2: %e, ss1: %e, ss2: %e, dnfpp:%e"%(it, em_co2, self.co2_hold["sCO2"][it], z_co2, self.co2_hold["yCO2"], self.co2_hold["xCO2"], self.co2_hold["ss1"], ss2, self.co2_hold["dfnpp"][it]))
        self.conc["CO2"][yr] = self.co2_hold["xCO2"]

    def fill_one_row_conc(self, yr, avoid=None):
        """
        Fill in one row of concentrations in conc_dict

        Fill in a row of zeros in concentrations dictionary
        from prescribed concentrations. If some traces are
        to be avoided (typically CO2 calculated from emissions)
        a list of them can be sent as inputs

        Parameters
        ----------
        yr : int
          Year for which to add concentrations from prescribed
        avoid : list
             optional list of tracers for which not to read
             presecribed concentrations. Typically CO2 when CO2
             is calculated from emissions, while other compounds
             are read from prescribed data.
        """
        for tracer, value_dict in self.conc.items():
            if avoid and tracer in avoid:
                continue
            if tracer in self.conc_in:
                value_dict[yr] = self.conc_in[tracer][yr]
            else:
                value_dict[yr] = 0

    def add_row_of_zeros_conc(self, yr):
        """
        Fill in one row of concentrations in conc_dict

        Fill in a row of zeros in concentrations dictionary
        for a single year

        Parameters
        ----------
        yr : int
          Year for which to add zeros
        """
        for value_dict in self.conc.values():
            value_dict[yr] = 0

    def write_output_to_files(self, cfg, make_plot=False):
        """
        Write results to files after run

        Write results for emissions, concentrations
        and forcings to files. Format is as in the
        fortran implementation, with separate files for
        emissions (emis), concentrations (conc) and
        forcing (forc)

        Parameters
        ----------
        cfg : dict
           Configurations to define where to put output
           files and what prefix to have for file name
        make_plot : bool
           Whether the output should be plottet or not
        """
        if "output_folder" in cfg:
            # Make os independent?
            outdir = os.path.join(os.getcwd(), cfg["output_folder"])
        else:
            outdir = os.path.join(os.getcwd(), "output")

        df_forc = pd.DataFrame(data=self.forc, index=self.years)
        df_forc["Year"] = self.years

        cols = df_forc.columns.tolist()
        cols = cols[-1:] + cols[:-1]
        df_forc = df_forc[cols]
        df_forc.rename(columns={"SO2": "SO4_DIR"}, inplace=True)
        df_emis = self.emis.drop(labels=["CO2_FF", "CO2_AFOLU"], axis=1).drop(
            labels=np.arange(self.years[-1] + 1, self.emis.index[-1] + 1), axis=0
        )
        df_emis["Year"] = self.years
        df_emis["CO2"] = self.emis["CO2_FF"] + self.emis["CO2_AFOLU"]
        cols = df_emis.columns.tolist()
        cols = cols[-2:] + cols[:-2]
        df_emis = df_emis[cols]
        self.conc["Year"] = self.years

        df_conc = pd.DataFrame(data=self.conc, index=self.years)
        cols = df_conc.columns.tolist()
        cols = cols[-1:] + cols[:-1]
        df_conc = df_conc[cols]
        for tracer in self.df_gas.index:
            if tracer not in df_emis.columns.tolist():
                df_emis[tracer] = np.zeros(len(self.years))
        if "output_prefix" in cfg:
            filename_start = cfg["output_prefix"]
        else:
            filename_start = "output"
        df_forc.to_csv(
            os.path.join(outdir, f"{filename_start}_forc.txt"),
            sep="\t",
            index=False,
            float_format="%.5e",
        )
        df_conc.to_csv(
            os.path.join(outdir, f"{filename_start}_conc.txt"),
            sep="\t",
            index=False,
            float_format="%.5e",
        )

        df_emis.to_csv(
            os.path.join(outdir, f"{filename_start}_em.txt"),
            sep="\t",
            index=False,
            float_format="%.5e",
        )

        if make_plot:
            plot_output2("forc", df_forc, outdir)
            plot_output2("emis", df_emis, outdir, self.df_gas["EM_UNIT"])
            plot_output2("conc", df_conc, outdir, self.df_gas["CONC_UNIT"])

    def add_results_to_dict(self):
        """
        Adding results to results dictionary

        Returns
        -------
        dict
            Containing run emissions, concentrations and forcings
            in the form of pandas.Dataframe with years and tracers
        """
        df_forc = pd.DataFrame(data=self.forc, index=self.years)
        df_forc["Year"] = self.years

        cols = df_forc.columns.tolist()
        cols = cols[-1:] + cols[:-1]
        df_forc = df_forc[cols]
        df_forc.rename(columns={"SO2": "SO4_DIR"}, inplace=True)
        df_emis = self.emis.drop(labels=["CO2_FF", "CO2_AFOLU"], axis=1).drop(
            labels=np.arange(self.years[-1] + 1, self.emis.index[-1] + 1), axis=0
        )
        df_emis["Year"] = self.years
        df_emis["CO2"] = self.emis["CO2_FF"] + self.emis["CO2_AFOLU"]
        cols = df_emis.columns.tolist()
        cols = cols[-2:] + cols[:-2]
        df_emis = df_emis[cols]
        self.conc["Year"] = self.years

        df_conc = pd.DataFrame(data=self.conc, index=self.years)
        cols = df_conc.columns.tolist()
        cols = cols[-1:] + cols[:-1]
        df_conc = df_conc[cols]
        results = {}
        results["emissions"] = df_emis
        results["concentrations"] = df_conc
        results["forcing"] = df_forc
        return results



#Plot radiative imbalance
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


#Read results:
outdir = '/div/no-backup/users/ragnhibs/ciceroscm/scripts/output_test/'
scen = 'test'
df_rib=pd.read_csv(outdir+'/output_rib.txt', sep='\t', index_col=0)
print(df_rib)


fig, axs = plt.subplots(nrows=1, ncols=2,sharex=True,figsize=(12,6))
axs=axs.flatten()
fig.suptitle('CICERO SCM simulation, RIB')


comp = 'RIB_glob'
df_rib[comp].plot(ylabel='check',ax=axs[0],label=scen)
axs[0].set_title(comp)



df_rib['RIB_glob'].plot(ylabel='check ',ax=axs[1],label=scen)
df_rib['RIB_N'].plot(ax=axs[1],linestyle=':',label='RIB_N')
df_rib['RIB_S'].plot(ax=axs[1],linestyle='--',label='RIB_S')
axs[1].set_title(comp)

axs[0].legend()
axs[1].legend()

plt.show()



exit()



import os
import sys

# Adding location of source code to system path
# os.path.dirname(__file__) gives the directory of
# current file. Put in updated path if running script from elsewhere
# os.path joins all the folders of a path together in a
# system independent way (i.e. will work equally well on Windows, linux etc)
sys.path.append(os.path.join(os.path.dirname(__file__), "../", "src"))

from ciceroscm import CICEROSCM

data_dir = os.path.join(os.path.dirname(__file__), "../", "tests", "test-data")
pert_dir = os.path.join(os.path.dirname(__file__), "pre_script")
outdir = os.path.join(os.getcwd(), "output_test")

# Emission perturbation test:
cscm = CICEROSCM(
    {
        "gaspam_file": os.path.join(data_dir, "gases_v1RCMIP.txt"),
        "nystart": 1900,
        "emstart": 1950,
        "nyend": 2050,
        "concentrations_file": os.path.join(data_dir, "ssp245_conc_RCMIP.txt"),
        "emissions_file": os.path.join(data_dir, "ssp245_em_RCMIP.txt"),
        "nat_ch4_file": os.path.join(data_dir, "natemis_ch4.txt"),
        "nat_n2o_file": os.path.join(data_dir, "natemis_n2o.txt"),
        "perturb_em_file": os.path.join(pert_dir, "pertem_test.txt"),        
    },
)

cscm._run({"output_folder": outdir, "output_prefix": "emissions_perturbation"})

# Forcing perturbation test:
cscm = CICEROSCM(
    {
        "gaspam_file": os.path.join(data_dir, "gases_v1RCMIP.txt"),
        "nystart": 1900,
        "emstart": 1950,
        "nyend": 2050,
        "concentrations_file": os.path.join(data_dir, "ssp245_conc_RCMIP.txt"),
        "emissions_file": os.path.join(data_dir, "ssp245_em_RCMIP.txt"),
        "nat_ch4_file": os.path.join(data_dir, "natemis_ch4.txt"),
        "nat_n2o_file": os.path.join(data_dir, "natemis_n2o.txt"),
        "perturb_forc_file": os.path.join(pert_dir, "pertforc_test.txt"),        
    },
)

cscm._run({"output_folder": outdir, "output_prefix": "forcing_perturbation"})



"""
CICEROSCM
"""

import logging
import os

import numpy as np
import pandas as pd

from ._utils import cut_and_check_pamset
from .concentrations_emissions_handler import ConcentrationsEmissionsHandler
from .input_handler import InputHandler
from .make_plots import plot_output1
from .upwelling_diffusion_model import UpwellingDiffusionModel

LOGGER = logging.getLogger(__name__)


class CICEROSCM:
    """
    Main ciceroscm class

    Takes care of and routes calls to diffusion
    model and concentrations emissions handler
    Takes results, and should be the module
    interacted with from other programs

    Attributes
    ----------
    cfg : dict
          Configurations dictionary with startyear,
          endyear, emissions start plus optional
          configs, like a forcing file path for pure
          forcing runs, else input files needed for
          concentrations and emission handler,
          a parameter to include volcanic and solar
          forcing, one to indicate a pure concentrations
          run, perturbation files etc.
    ce_handler : ciceroscm.ConcentrationsEmissionsHandler
                     Concentrations emissions handler for the
                     model. It can be reset for multiple runs
                     but needs to be used with the same year range
                     and with either Concentrations or emissions
                     run consistently
    results : dict
                  Results dictionary which stores results from
                  upwelling diffusion model. If a results_as_dict
                  is sent as True to the run call, the results
                  from the Concentrations emissions handler
                  (emissions, componentwise forcings and concentrations)
                  will also be added to this dictionary and can be
                  accessed from there
    rf_volc_sun : dict
                  With solar and volcanic forcing read from standard files
                  or set to zero according to configurations
    rf_luc : pd.Dataframe
             Dataframe of land use albedo forcing with years as index
    """

    # pylint: disable=too-many-instance-attributes
    # Consider whether this can be cut back later

    def __init__(self, cfg):
        """
        Intialise CICEROSCM

        Starting by picking out the part of cfg that
        are needed as the class cfgs.
        Then we either read in forc_file for forcing
        run, or check file availability and set start,
        end and emissions start for concentrations or
        emissions run.
        Then make empty dictionary for results, read in
        solar and volcanic forcing and initialise other
        output arrays

        Parameters
        ----------
        cfg : dict
              Configurations containing inputs about class
              such as a forcing file for forcing run,
              locations of files to use for concentration
              or emission runs, and start and end of run etc.

        """
        self.cfg = cut_and_check_pamset(
            {"nystart": 1750, "nyend": 2100, "emstart": 1850, "idtm": 24}, cfg
        )
        cfg.update(self.cfg)
        input_handler = InputHandler(cfg)
        self.cfg["rf_run"] = input_handler.optional_pam("forc")
        if self.cfg["rf_run"]:
            self.rf = input_handler.get_data("forc")
        else:
            # cfg = check_inputfiles(cfg)
            pamset_emiconc = {}
            pamset_emiconc["emstart"] = self.cfg["emstart"]
            pamset_emiconc["nystart"] = self.cfg["nystart"]
            pamset_emiconc["nyend"] = self.cfg["nyend"]
            pamset_emiconc["idtm"] = self.cfg["idtm"]
            for key, value in cfg.items():
                if key in ["rs_function", "rb_function"]:
                    pamset_emiconc[key] = value
            self.ce_handler = ConcentrationsEmissionsHandler(
                input_handler, pamset_emiconc
            )
        self.results = {}
        # Reading in solar and volcanic forcing
        self.rf_volc_sun = {
            "volc_n": input_handler.get_data("rf_volc_n"),
            "volc_s": input_handler.get_data("rf_volc_s"),
            "sun": input_handler.get_data("rf_sun"),
        }

        # Add support for sending filename in cfg
        self.rf_luc = input_handler.get_data("rf_luc")
        self.initialise_output_arrays()

    def initialise_output_arrays(self):
        """
        Initialise dict with arrays to hold data for run

        Dictionary for all results from upwelling diffusion
        model outputs is initialised with empty arrays
        """
        output_variables = [
            "OHC700",
            "OHCTOT",
            "RIB_glob",
            "RIB_N",
            "RIB_S",
            "dT_glob",
            "dT_NH",
            "dT_SH",
            "dT_glob_air",
            "dT_NH_air",
            "dT_SH_air",
            "dT_glob_sea",
            "dT_NH_sea",
            "dT_SHsea",
            "Total_forcing",
            "Solar_forcing",
            "Volcanic_forcing_NH",
            "Volcanic_forcing_SH",
        ]
        for output in output_variables:
            self.results[output] = np.zeros(self.cfg["nyend"] - self.cfg["nystart"] + 1)

    def forc_set(self, yr, rf_sun):
        """
        Read the forcing for this year

        Getting a year, getting the forcung for this year and
        adding solar forcing.

        Parameters
        ----------
        yr : int
          Year for which to read out data
        rf_sun : pandas.Dataframe
              Dataframe with solar forcing to  add to the other
              forcings.

        Returns
        -------
        float
             The total forcing for the year, including solar
             forcing is added.
        """
        row_index = yr - self.cfg["nystart"]
        # Add support for other forcing formats
        if isinstance(self.rf, np.ndarray):
            # Add luc albedo later
            forc = self.rf[row_index]  # + self.rf_luc.iloc[row_index][0]
            fn = forc
            fs = forc
        else:
            forc = self.rf["total"][yr]
            fn = self.rf["FORC_NH"][yr]
            fs = self.rf["FORC_SH"][yr]
        forc = forc + rf_sun.iloc[row_index, 0]
        fn = fn + rf_sun.iloc[row_index, 0]
        fs = fs + rf_sun.iloc[row_index, 0]
        return fn, fs, forc

    def add_year_data_to_output(self, values, forc, index):
        """
        Add single year output to output arrays

        Add all the outputs from a single year run
        of upwelling diffusion model to output arrays

        Parameters
        ----------
        values : dict
              Output from upwelling diffusion model
        forc : float
            Total forcing for this year
        index : int
             Index equalling year number in the possible years
        """
        simple_outputs = ["OHC700", "OHCTOT"]
        for output in simple_outputs:
            self.results[output][index] = values[output]
        outputs_dict = {
            "RIB_glob": "RIB",
            "RIB_N": "RIBN",
            "RIB_S": "RIBS",
            "dT_glob": "dtemp",
            "dT_NH": "dtempnh",
            "dT_SH": "dtempsh",
            "dT_glob_air": "dtemp_air",
            "dT_NH_air": "dtempnh_air",
            "dT_SH_air": "dtempsh_air",
            "dT_glob_sea": "dtemp_sea",
            "dT_NH_sea": "dtempnh_sea",
            "dT_SHsea": "dtempsh_sea",
        }
        for output, name in outputs_dict.items():
            self.results[output][index] = values[name]
        self.results["Total_forcing"][index] = forc
        self.results["Solar_forcing"][index] = self.rf_volc_sun["sun"].iloc[index, 0]
        self.results["Volcanic_forcing_NH"][index] = np.mean(
            np.array(self.rf_volc_sun["volc_n"].iloc[index, :])
        )
        self.results["Volcanic_forcing_SH"][index] = np.mean(
            np.array(self.rf_volc_sun["volc_s"].iloc[index, :])
        )

    def _run(
        self, cfg, pamset_udm={}, pamset_emiconc={}, make_plot=False
    ):  # pylint: disable=dangerous-default-value
        """
        Run CICEROSCM

        Setting off a full model run. Starting by
        intialising output arrays, and udm_model and
        resetting ConcentrationEmissionsHandler for a new run
        Then looping over year by year converting emissions
        and concnetrations to forcings if applicable
        and then running the upwelling diffusion model
        Finally writing results to file

        Parameters
        ----------
        cfg : dict
           Dictionary with run specific configurations
        pamset_udm : dict
                  Parameter set for udm model
        pamset_emiconc : dict
                      Parameter set for concentrations
                      emissions handler
        """
        self.initialise_output_arrays()
        # Setting up UDM
        udm = UpwellingDiffusionModel(pamset_udm)
        if not self.cfg["rf_run"]:
            pamset_emiconc["emstart"] = self.cfg["emstart"]
            pamset_emiconc["nystart"] = self.cfg["nystart"]
            pamset_emiconc["nyend"] = self.cfg["nyend"]
            self.ce_handler.reset_with_new_pams(pamset_emiconc)
        for yr in range(self.cfg["nystart"], self.cfg["nyend"] + 1):
            if not self.cfg["rf_run"]:
                self.ce_handler.emi2conc(yr)
                forc, fn, fs = self.ce_handler.conc2forc(
                    yr,
                    self.rf_luc.iloc[yr - self.cfg["nystart"], 0],
                    self.rf_volc_sun["sun"].iloc[yr - self.cfg["nystart"], 0],
                )

            else:
                fn, fs, forc = self.forc_set(yr, self.rf_volc_sun["sun"])
            values = udm.energy_budget(
                fn,
                fs,
                np.array(self.rf_volc_sun["volc_n"].iloc[yr - self.cfg["nystart"], :]),
                np.array(self.rf_volc_sun["volc_s"].iloc[yr - self.cfg["nystart"], :]),
            )
            self.add_year_data_to_output(values, forc, yr - self.cfg["nystart"])

        if make_plot:
            plot_output1(cfg, self.results, self.cfg["nystart"], self.cfg["nyend"])
        if ("results_as_dict" in cfg) and cfg["results_as_dict"]:
            if not self.cfg["rf_run"]:
                self.results.update(self.ce_handler.add_results_to_dict())
        else:
            if not self.cfg["rf_run"]:
                self.ce_handler.write_output_to_files(cfg, make_plot)

            self.write_data_to_file(cfg)

    def write_data_to_file(self, pamset):
        """
        Write results to files after run

        Writing results from upwelling diffusion model to file
        Formatting and organising in ocean heat content (ohc),
        radiative imbalance (rib), temperature related (temp),
        and forcing (forc) files are as in original fortran
        implementation. Forcing file is only outputted here
        if the run is a forcing run. Otherwise the forcing results
        writing is handled by the ConcentrationsEmissionsHandler

        Parameters
        ----------
        pamset : dict
              parameterset with details on where to write results

        """
        if "output_folder" in pamset:
            # Make os independent?
            outdir = os.path.join(os.getcwd(), pamset["output_folder"])
        else:
            outdir = os.path.join(os.getcwd(), "output")

        indices = np.arange(self.cfg["nystart"], self.cfg["nyend"] + 1)
        df_ohc = pd.DataFrame(
            data={
                "Year": indices,
                "OHC700": self.results["OHC700"],
                "OHCTOT": self.results["OHCTOT"],
            }
        )
        list_rib = ["RIB_glob", "RIB_N", "RIB_S"]
        df_rib = pd.DataFrame(data={"Year": indices})
        for vari in list_rib:
            df_rib[vari] = self.results[vari]
        list_temp = [
            "dT_glob",
            "dT_NH",
            "dT_SH",
            "dT_glob_air",
            "dT_NH_air",
            "dT_SH_air",
            "dT_glob_sea",
            "dT_NH_sea",
            "dT_SHsea",
        ]
        df_temp = pd.DataFrame(data={"Year": indices})
        if "output_prefix" in pamset:
            filename_start = pamset["output_prefix"]
        else:
            filename_start = "output"
        for vari in list_temp:
            df_temp[vari] = self.results[vari]

        df_ohc.to_csv(
            os.path.join(outdir, f"{filename_start}_ohc.txt"),
            sep="\t",
            index=False,
            float_format="%.5e",
        )
        df_rib.to_csv(
            os.path.join(outdir, f"{filename_start}_rib.txt"),
            sep="\t",
            index=False,
            float_format="%.5e",
        )
        df_temp.to_csv(
            os.path.join(outdir, f"{filename_start}_temp.txt"),
            sep="\t",
            index=False,
            float_format="%.5e",
        )
        df_sunvolc = pd.DataFrame(
            data={
                "Year": indices,
                "Solar_forcing": self.results["Solar_forcing"],
                "Volcanic_forcing_NH": self.results["Volcanic_forcing_NH"],
                "Volcanic_forcing_SH": self.results["Volcanic_forcing_SH"],
            }
        )
        df_sunvolc.to_csv(
            os.path.join(outdir, f"{filename_start}_sunvolc.txt"),
            sep="\t",
            index=False,
            float_format="%.5e",
        )
        if self.cfg["rf_run"]:
            df_forc = pd.DataFrame(
                data={"Year": indices, "Total_forcing": self.results["Total_forcing"]}
            )
            df_forc.to_csv(
                os.path.join(outdir, f"{filename_start}_forc.txt"),
                sep="\t",
                index=False,
                float_format="%.5e",
            )



import os

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

import versioneer

PACKAGE_NAME = "ciceroscm"
AUTHORS = [
    ("Marit Sandstad", "marit.sandstad@cicero.oslo.no"),
    ("Ragnhild Bieltvedt Skeie", "r.b.skeie@cicero.oslo.no"),
    ("Ane Nordlie Johansen", "ane.nordlie.johansen@cicero.oslo.no"),
    ("Benjamin Sanderson", "benjamin.sanderson@cicero.oslo.no"),
]
URL = "https://github.com/ciceroOslo/ciceroscm"

DESCRIPTION = "Python version of the CICERO-SCM simple climate model"
README = "README.rst"

SOURCE_DIR = "src"

REQUIREMENTS = [
    "click",
    "openscm-units>=0.5.0",
    "pyam-iamc",
    "python-dotenv",
    "scmdata>=0.7.4",
    "tqdm",
    "matplotlib>=3.4",
    "openscm_runner",
]

REQUIREMENTS_NOTEBOOKS = [
    "ipywidgets",
    "notebook",
    "seaborn",
]
REQUIREMENTS_TESTS = [
    "codecov",
    "coverage",
    "nbval",
    "pytest-cov",
    "pytest>=4.0",
    "xlrd",
]
REQUIREMENTS_DOCS = ["sphinx>=1.4", "sphinx_rtd_theme", "sphinx-click"]
REQUIREMENTS_DEPLOY = ["twine>=1.11.0", "setuptools>=41.2", "wheel>=0.31.0"]

REQUIREMENTS_DEV = [
    *[
        "bandit",
        "black>=22.3.0",
        "black-nb",
        "flake8",
        "isort>5",
        "mypy",
        "nbdime",
        "pydocstyle",
        "pylint>=2.4.4",
    ],
    *REQUIREMENTS_DEPLOY,
    *REQUIREMENTS_DOCS,
    *REQUIREMENTS_NOTEBOOKS,
    *REQUIREMENTS_TESTS,
]

REQUIREMENTS_EXTRAS = {
    "deploy": REQUIREMENTS_DEPLOY,
    "dev": REQUIREMENTS_DEV,
    "docs": REQUIREMENTS_DOCS,
    "notebooks": REQUIREMENTS_NOTEBOOKS,
    "tests": REQUIREMENTS_TESTS,
}

# no tests/docs in `src` so don't need exclude
PACKAGES = find_packages(SOURCE_DIR)
PACKAGE_DIR = {"": SOURCE_DIR}
PACKAGE_DATA = {"ciceroscm": [os.path.join("default_data", "*.txt")]}

# Get the long description from the README file
with open(README, "r") as f:
    README_LINES = ["ciceroscm", "==============", ""]
    add_line = False
    for line in f:
        if line.strip() == ".. sec-begin-long-description":
            add_line = True
        elif line.strip() == ".. sec-end-long-description":
            break
        elif add_line:
            README_LINES.append(line.strip())

if len(README_LINES) < 3:
    raise RuntimeError("Insufficient description given")


class ciceroscm(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": ciceroscm})

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description="\n".join(README_LINES),
    long_description_content_type="text/x-rst",
    author=", ".join([author[0] for author in AUTHORS]),
    author_email=", ".join([author[1] for author in AUTHORS]),
    url=URL,
    license="Apache 2.0",
    classifiers=[  # full list at https://pypi.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords=["cicero-scm", "cicero", "python", "repo", "simple", "climate", "model"],
    packages=PACKAGES,
    package_dir=PACKAGE_DIR,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    install_requires=REQUIREMENTS,
    extras_require=REQUIREMENTS_EXTRAS,
    cmdclass=cmdclass,
)



import os

from ciceroscm import CICEROSCM


def check_output(
    output_dir,
    expected_output_dir,
    files=[
        "conc_1.png",
        "conc_2.png",
        "em_1.png",
        "em_2.png",
        "em_3.png",
        "forc_1.png",
        "forc_2.png",
        "forc_3.png",
        "ohc.png",
        "rib.png",
        "temp.png",
    ],
):
    for filename in files:
        file_expected = os.path.join(expected_output_dir, "plots", filename)
        assert os.path.exists(file_expected.replace(expected_output_dir, output_dir))


def test_plot(tmpdir, test_data_dir):
    cscm = CICEROSCM(
        {
            "gaspam_file": os.path.join(test_data_dir, "gases_v1RCMIP.txt"),
            "nyend": 2100,
            "nystart": 1750,
            "emstart": 1850,
            "concentrations_file": os.path.join(test_data_dir, "ssp245_conc_RCMIP.txt"),
            "emissions_file": os.path.join(test_data_dir, "ssp245_em_RCMIP.txt"),
            "nat_ch4_file": os.path.join(test_data_dir, "natemis_ch4.txt"),
            "nat_n2o_file": os.path.join(test_data_dir, "natemis_n2o.txt"),
        },
    )

    outdir = str(tmpdir)

    cscm._run({"output_folder": outdir}, make_plot=True)

    check_output(outdir, os.path.join(test_data_dir, "ssp245_emis"))



"""
Class for common utility functions
"""

import logging

LOGGER = logging.getLogger(__name__)


def check_numeric_pamset(required, pamset):
    """
    Check numeric pamset conforms

    Check that pamset has required inputs
    If not insert defaults from required dict

    Parameters
    ----------
    required : dict
            Dictionary containing the required keys
            with default values
    pamset : dict
          Dictionary to be checked and filled in from
          required where necessary

    Returns
    -------
    dict
         pamset augmented with necessary default values
         from required
    """
    for pam, value in required.items():
        if pam not in pamset:
            LOGGER.warning(  # pylint: disable=logging-fstring-interpolation
                f"Parameter {pam} not in pamset. Using default value {value}",
            )
            pamset[pam] = value
        elif not isinstance(pamset[pam], int) and not isinstance(pamset[pam], float):
            LOGGER.warning(  # pylint: disable=logging-fstring-interpolation
                f"Parameter {pam} must be a number. Using default value {value}",
            )
            pamset[pam] = value
    return pamset


def cut_non_required(required, pamset, cut_warnings=False):
    """
    Cut elements from pamset that are not required

    Take out elements of pamset that are not used

    Parameters
    ----------
    required : dict
            Dictionary containing the required keys
            with default values
    pamset : dict
          Dictionary to be checked and cut from
    cut_warnings : bool
                Boolean stating whether or not warnings
                should be printed if parameters are cut
    Returns
    -------
    dict
         pamset where unnecessary elements have been cut
    """
    new_pamset = {}
    for pam in pamset:
        if pam in required:
            new_pamset[pam] = pamset[pam]
        elif cut_warnings:
            LOGGER.warning(  # pylint: disable=logging-fstring-interpolation
                f"Parameter {pam} is not used. Please check if you have a typo",
            )
    return new_pamset


def cut_and_check_pamset(
    required, pamset, used={}, cut_warnings=False
):  # pylint: disable=dangerous-default-value
    """
    Combine cut and check of parameters

    Calling other methods to first cut unnecessary
    elements and afterwords put in missing elements

    Parameters
    ----------
    required : dict
            Dictionary containing the required keys
            with default values
    pamset : dict
          Dictionary to be checked and cut from
    used : dict
        Dictionary or list containing keys used, so
        that unused parameters can be cut.
    cut_warnings : bool
                Boolean stating whether or not warnings
                should be printed if parameters are cut
    Returns
    -------
    dict
         pamset where unnecessary elements have been cut
         and missing necessary elements have been added
         or augmented if not of numeric type.
    """
    if used:
        used.update(required)
    else:
        used = required
    pamset = cut_non_required(used, pamset, cut_warnings)
    return check_numeric_pamset(required, pamset)



import os

import numpy as np

from ciceroscm import input_handler
from ciceroscm.parallel.cscmparwrapper import run_ciceroscm_parallel


def test_ciceroscm_run_parallel_many_scenarios(test_data_dir):
    scenarios = []
    gaspam_data = input_handler.read_components(
        os.path.join(test_data_dir, "gases_v1RCMIP.txt")
    )
    conc_data = input_handler.read_inputfile(
        os.path.join(test_data_dir, "ssp245_conc_RCMIP.txt"), True, 1750, 2100
    )
    ih = input_handler.InputHandler({"nyend": 2050, "nystart": 1900, "emstart": 2015})
    em_data = ih.read_emissions(os.path.join(test_data_dir, "ssp245_em_RCMIP.txt"))
    nat_ch4_data = input_handler.read_natural_emissions(
        os.path.join(test_data_dir, "natemis_ch4.txt"), "CH4"
    )
    nat_n2o_data = input_handler.read_natural_emissions(
        os.path.join(test_data_dir, "natemis_n2o.txt"), "N2O"
    )
    cfgs = [
        {
            "pamset_udm": {
                "rlamdo": 15.1,
                "akapa": 0.657,
                "cpi": 0.208,
                "W": 2.2,
                "beto": 6.9,
                "lambda": 0.606,
                "mixed": 107.0,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.5,
                "qdirso2": -0.3701,
                "qindso2": -0.4163,
                "qbc": 0.163,
                "qoc": -0.084,
                "qh2o_ch4": 0.171,
            },
            "Index": "13555_old_NR_rounded",
        }
    ]
    for i in range(20):
        em_data_here = em_data.mul((1 + i / 100))
        new_scen = {
            "gaspam_data": gaspam_data,
            "nyend": 2050,
            "nystart": 1900,
            "emstart": 2015,
            "concentrations_data": conc_data,
            "nat_ch4_data": nat_ch4_data,
            "nat_n2o_data": nat_n2o_data,
            "emissions_data": em_data_here,
            "udir": test_data_dir,
            "scenname": "ssp245-plus-%d-percent" % i,
        }
        scenarios.append(new_scen)
    output_variables = ["Heat Content|Ocean", "Surface Air Temperature Change"]
    results = run_ciceroscm_parallel(scenarios, cfgs, output_variables)
    print(results)
    assert set(results.get_unique_meta("variable")) == set(
        ["Heat Content|Ocean", "Surface Air Temperature Change"]
    )
    assert set(results.get_unique_meta("scenario")) == set(
        [f"ssp245-plus-{d}-percent" for d in range(20)]
    )
    test_length = results.filter(
        variable="Surface Air Temperature Change",
        scenario="ssp245-plus-1-percent",
        run_id="13555_old_NR_rounded",
    )
    test_not_empty = results.filter(year=2000)
    assert len(test_length.values[0]) == 151
    assert test_not_empty.values.any()


def test_ciceroscm_run_parallel_many_cfgs(test_data_dir):
    gaspam_data = input_handler.read_components(
        os.path.join(test_data_dir, "gases_v1RCMIP.txt")
    )
    conc_data = input_handler.read_inputfile(
        os.path.join(test_data_dir, "ssp245_conc_RCMIP.txt"), True, 1750, 2100
    )
    ih = input_handler.InputHandler({"nyend": 2050, "nystart": 1900, "emstart": 2015})
    em_data = ih.read_emissions(os.path.join(test_data_dir, "ssp245_em_RCMIP.txt"))
    nat_ch4_data = input_handler.read_natural_emissions(
        os.path.join(test_data_dir, "natemis_ch4.txt"), "CH4"
    )
    nat_n2o_data = input_handler.read_natural_emissions(
        os.path.join(test_data_dir, "natemis_n2o.txt"), "N2O"
    )
    scenarios = [
        {
            "gaspam_data": gaspam_data,
            "nyend": 2050,
            "nystart": 1900,
            "emstart": 2015,
            "concentrations_data": conc_data,
            "nat_ch4_data": nat_ch4_data,
            "nat_n2o_data": nat_n2o_data,
            "emissions_data": em_data,
            "udir": test_data_dir,
            "scenname": "ssp245",
        }
    ]
    cfgs = [
        {
            "pamset_udm": {
                "rlamdo": 15.1,
                "akapa": 0.657,
                "cpi": 0.208,
                "W": 2.2,
                "beto": 6.9,
                "lambda": 0.606,
                "mixed": 107.0,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.5,
                "qdirso2": -0.3701,
                "qindso2": -0.4163,
                "qbc": 0.163,
                "qoc": -0.084,
                "qh2o_ch4": 0.171,
            },
            "Index": "13555_old_NR_rounded",
        },
        {
            "pamset_udm": {
                "rlamdo": 15.08357,
                "akapa": 0.6568376339229769,
                "cpi": 0.2077266,
                "W": 2.205919,
                "beto": 6.89822,
                "lambda": 0.6062529,
                "mixed": 107.2422,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.5,
                "qdirso2": -0.3562,
                "qindso2": -0.96609,
                "qbc": 0.1566,
                "qoc": -0.0806,
            },
            "Index": "13555_old_NR_improved",
        },
        {
            "pamset_udm": {
                "rlamdo": 15.08357,
                "akapa": 0.6568376339229769,
                "cpi": 0.2077266,
                "W": 2.205919,
                "beto": 6.89822,
                "lambda": 0.6062529,
                "mixed": 107.2422,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.5,
                "qdirso2": -0.3701323367808028 / 1.5,
                "qindso2": -0.4162980444986502 * 1.5,
                "qbc": 0.162692563111132,
                "qoc": -0.08377713183167902,
            },
            "Index": "13555_old_NR",
        },
        {
            "pamset_udm": {
                "rlamdo": 5.269455,
                "akapa": 0.40099950002568496,
                "cpi": 0.2054687,
                "W": 1.95183,
                "beto": 3.278654,
                "lambda": 0.7308369,
                "mixed": 115.1219,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.3,
                "qdirso2": -0.32211222516087934,
                "qindso2": -0.3622885009449893,
                "qbc": 0.1415852070009452,
                "qoc": -0.07290808089507649,
            },
            "Index": "10496_old_NR",
        },
        {
            "pamset_udm": {
                "rlamdo": 5.269455,
                "akapa": 0.40099950002568496,
                "cpi": 0.2054687,
                "W": 1.95183,
                "beto": 3.278654,
                "lambda": 0.7308369,
                "mixed": 115.1219,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.3,
                "qdirso2": -0.32211222516087934,
                "qindso2": -0.3622885009449893,
                "qbc": 0.1415852070009452,
                "qoc": -0.07290808089507649,
            },
            "Index": "10974_old_NR",
        },
        {
            "pamset_udm": {
                "rlamdo": 23.71469,
                "akapa": 0.5486541129586187,
                "cpi": 0.4059296,
                "W": 2.090927,
                "beto": 4.426507,
                "lambda": 0.6345007,
                "mixed": 97.25478,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.3,
                "qdirso2": -0.4031202924043847,
                "qindso2": -0.453400507735301,
                "qbc": 0.17719249872571508,
                "qoc": -0.09124374858602939,
            },
            "Index": "Unknown_old_NR",
        },
        {
            "pamset_udm": {
                "rlamdo": 23.13088,
                "akapa": 0.6588532950589802,
                "cpi": 0.1690946,
                "W": 2.228695,
                "beto": 2.623041,
                "lambda": 0.5402487,
                "mixed": 99.86714,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.3,
                "qdirso2": -0.882827534329248,
                "qindso2": -0.9929404692583796,
                "qbc": 0.3880489761967636,
                "qoc": -0.1998224726091362,
            },
            "Index": "28925_old_NR",
        },
    ]

    output_variables = ["Heat Content|Ocean", "Surface Air Temperature Change"]
    results = run_ciceroscm_parallel(scenarios, cfgs, output_variables)
    print(results)
    assert set(results.get_unique_meta("run_id")) == set(
        [
            "13555_old_NR_rounded",
            "13555_old_NR_improved",
            "13555_old_NR",
            "10496_old_NR",
            "10974_old_NR",
            "Unknown_old_NR",
            "28925_old_NR",
        ]
    )


def test_ciceroscm_run_parallel_many_forcing(test_data_dir):
    gaspam_data = input_handler.read_components(
        os.path.join(test_data_dir, "gases_v1RCMIP.txt")
    )
    scenarios = [
        {
            "gaspamfile": gaspam_data,
            "nyend": 2100,
            "forc_data": np.loadtxt(os.path.join(test_data_dir, "test_forcing.txt")),
            "udir": test_data_dir,
            "scenname": "forc_data_test_forcing",
        },
        {
            "gaspamfile": gaspam_data,
            "nyend": 2100,
            "forc_data": np.loadtxt(os.path.join(test_data_dir, "zero_forcing.txt")),
            "udir": test_data_dir,
            "scenname": "forc_data_zero_forcing",
        },
        {
            "gaspamfile": gaspam_data,
            "nyend": 2100,
            "forc_data": np.loadtxt(os.path.join(test_data_dir, "test_forcing.txt")),
            "sunvolc": 1,
            "udir": test_data_dir,
            "scenname": "forc_data_test_forcing_sunvolc",
        },
        {
            "gaspamfile": gaspam_data,
            "nyend": 2100,
            "forc_data": np.loadtxt(os.path.join(test_data_dir, "CO2_1pros.txt")),
            "udir": test_data_dir,
            "scenname": "forc_data_CO2_1pros",
        },
        {
            "gaspamfile": gaspam_data,
            "nyend": 2100,
            "forc_file": os.path.join(test_data_dir, "test_forcing_hemisplit.txt"),
            "udir": test_data_dir,
            "scenname": "forc_data_test_forcing_hemisplit",
        },
    ]
    cfgs = [
        {
            "pamset_udm": {
                "rlamdo": 15.1,
                "akapa": 0.657,
                "cpi": 0.208,
                "W": 2.2,
                "beto": 6.9,
                "lambda": 0.606,
                "mixed": 107.0,
            },
            "pamset_emiconc": {
                "qbmb": 0.0,
                "qo3": 0.5,
                "qdirso2": -0.3701,
                "qindso2": -0.4163,
                "qbc": 0.163,
                "qoc": -0.084,
                "qh2o_ch4": 0.171,
            },
            "Index": "10496_old_NR_rounded",
        }
    ]
    output_variables = ["Heat Content|Ocean", "Surface Air Temperature Change"]
    results = run_ciceroscm_parallel(scenarios, cfgs, output_variables)
    print(results)
    assert set(results.get_unique_meta("scenario")) == set(
        [
            "forc_data_test_forcing_hemisplit",
            "forc_data_CO2_1pros",
            "forc_data_test_forcing_sunvolc",
            "forc_data_zero_forcing",
            "forc_data_test_forcing",
        ]
    )
    test_length = results.filter(
        variable="Surface Air Temperature Change",
        scenario="forc_data_zero_forcing",
        run_id="10496_old_NR_rounded",
    )
    assert len(test_length.values[0]) == 351



"""
Energy budget upwelling diffusion model
"""

import logging

import numpy as np
from scipy.linalg import solve_banded

from ._utils import cut_and_check_pamset

SEC_DAY = 86400
DAY_YEAR = 365.0

LOGGER = logging.getLogger(__name__)


def _band(a_array, b_array, c_array, d_array):
    """
    Calculate band

    Parameters
    ----------
    a_array : np.ndarray
              a_array through ocean layers
    b_array : np.ndarray
              b_array through ocean layers
    c_array : np.ndarray
              c_array through ocean layers
    d_array : np.ndarray
               d_array through ocean layers

    Returns
    -------
    np.ndarray
             band value through ocean layers
    """
    return solve_banded((1, 1), np.array([c_array, b_array, a_array]), d_array)


def check_pamset(pamset):
    """
    Check that parameterset has necessary values for run

    Check that parameterset has necessary values for run
    Otherwise set to default values which are defined here

    Parameters
    ----------
    pamset : dict
          Dictionary of parameters to define the physics
          of the run

    Returns
    -------
    dict
        Updated pamset with default values used where necessary
    """
    required = {
        "rlamdo": 15.0,
        "akapa": 0.66,
        "cpi": 0.21,
        "W": 2.2,
        "beto": 6.9,
        "threstemp": 7.0,
        "lambda": 0.61,
        "mixed": 107.0,
        "foan": 0.61,
        "foas": 0.81,
        "ebbeta": 0.0,
        "fnso": 0.7531,
        "lm": 40,
        "ldtime": 12,
    }
    pamset = cut_and_check_pamset(required, pamset, cut_warnings=True)
    pamset["rakapa"] = 1.0e-4 * pamset["akapa"]
    pamset["rlamda"] = 1.0 / pamset["lambda"]
    pamset["dt"] = 1 / pamset["ldtime"] * SEC_DAY * DAY_YEAR
    rho = 1.03
    htcpty = 0.955
    cnvrt = 0.485
    pamset["c1"] = rho * htcpty * cnvrt * 100.0 * SEC_DAY
    pamset["fnx"] = (
        pamset["rlamda"] + pamset["foan"] * pamset["rlamdo"] + pamset["ebbeta"]
    )
    pamset["fsx"] = (
        pamset["rlamda"] + pamset["foas"] * pamset["rlamdo"] + pamset["ebbeta"]
    )
    return pamset


class UpwellingDiffusionModel:  # pylint: disable=too-many-instance-attributes
    """
    Class to handle energy budget upwelling and downwelling

    Attributes
    ----------
    pamset : dict
             Dictionary of parameter values
    dz : np.ndarray
         Array of depth of ocean layers
    varrying : dict
               Dictionary of coefficient, temperature values
               etc which varry over time, but do not need to
               be stored apart from current value
    tn : np.ndarray
         Temperature change in the ocean layers of the
         Northern hemisphere
    ts : np.ndarray
         Temperature change in the ocean layers of the
         Southern hemisphere
    prev_values : dict
                  Dictionary with values from pervious time step
                  needed in calculations, typically previous
                  temperature change etc
    dtempprev : float
                Temperature change in previous time step
    press : np.ndarray
            Pressure in ocean layers
    tempunp : np.ndarray
              Temperature in ocean layers
    dens0 : np.ndarray
            Density in ocean layers
    """

    def __init__(self, params):
        """
        Intialise

        Setting up heights first, then starting up empty dict and
        arrays , and some starting values

        Parameters
        ----------
        params : dict
              Physical parameters to define the instance
        """
        self.pamset = check_pamset(params)

        # Setting up dz height difference between ocean layers
        self.dz = np.ones(self.pamset["lm"]) * 100.0
        self.dz[0] = self.pamset["mixed"]
        self.varrying = {}
        self.setup_ebud()

        # Intialising temperature values
        self.tn = np.zeros(self.pamset["lm"])
        self.ts = np.zeros(self.pamset["lm"])
        # Dict of values to keep from one year to the next
        self.prev_values = {
            "fn": 0.0,
            "fs": 0.0,
            "dtemp": 0.0,
        }

        self.dtempprev = 0.0

    def get_gam_and_fro_factor_ns(self, northern_hemisphere):
        """
        Get correct gam and fro variables

        Get correct gam and fro variables depending on
        whether Northern or Southern hemisphere is
        considered

        Parameters
        ----------
        northern_hemisphere : bool
                           Whether northern or southern hemisphere
                           is being considered

        Returns
        -------
        float
             The correct gam_fro_factor
        """
        blm = self.pamset["ebbeta"] / self.pamset["rlamdo"]
        if northern_hemisphere:
            gam1 = self.gamn
            gam2 = self.gams
            fro1 = self.pamset["foan"]
        else:
            gam1 = self.gams
            gam2 = self.gamn
            fro1 = self.pamset["foas"]
        factor = (
            self.pamset["rlamdo"]
            * (1.0 - fro1 * gam2 / (gam2 * gam1 - blm * blm))
            * self.pamset["dt"]
            / (self.pamset["c1"] * self.dz[0])
        )
        return factor

    def coeff(self, wcfac, gam_fro_fac):
        """
        Calculate a, b c coefficient arrays for hemisphere

        This method has been changed since the fortran version
        to have the coefficients a be the coefficients for the
        layer underneath, b be the coefficient for current layer
        and c be the coefficient for the layer above in such a
        way that they can represent a banded matrix and solved for
        as such

        Parameters
        ----------
        wcfac : float
             wc factor is W per fraction of the year
        gam_fro_fac : float
                      Is the factor combo of gam and fro
        Returns
        -------
        list
            Containing the a, b and c coefficients over the ocean layers
        """
        lm = self.pamset["lm"]
        a = np.zeros(lm)
        b = np.zeros(lm)
        c = np.zeros(lm)
        rakapafac = 2 * self.pamset["rakapa"] * self.pamset["dt"]

        c[1] = -rakapafac / (
            self.dz[0] * (0.0 * self.dz[0] + self.dz[1])
        )  # Can the 0.*dz(0) term be dropped here?
        b[0] = 1.0 - c[1] + gam_fro_fac - wcfac / self.dz[0]
        a[0] = -rakapafac / (self.dz[1] ** 2) + wcfac / self.dz[1]
        a[1 : lm - 1] = -rakapafac / (self.dz[2:] * (self.dz[1 : lm - 1] + self.dz[2:]))
        c[2:] = (
            -rakapafac / (self.dz[1 : lm - 1] * (self.dz[1 : lm - 1] + self.dz[2:]))
            - wcfac / self.dz[1 : lm - 1]
        )
        b[1 : lm - 1] = 1.0 - a[: lm - 2] - c[2:]
        b[lm - 1] = (
            1.0 - a[lm - 2] + wcfac / self.dz[lm - 1]
        )  # Her var det brukt i selvom vi var utenfor loekka, litt uklart hva som er ment...
        return a, b, c

    def setup_ebud2(self, temp_1n, temp_1s):
        """
        Set up coefficients and more for the two hemispheres

        Set up coefficients and various parameters
        for the two hemispheres. To be redone every timestep

        Parameters
        ----------
        temp_1n : float
               Northern hemisphere surface temperature
        temp_1s : float
               Southern hemisphere surface temperature
        """
        # Northern hemisphere:
        if self.pamset["threstemp"] == 0:  # pylint: disable=compare-to-zero
            wcfac = self.pamset["W"] / (SEC_DAY * DAY_YEAR) * self.pamset["dt"]
        else:
            wcfac = (
                self.pamset["W"]
                / (SEC_DAY * DAY_YEAR)
                * (1 - 0.3 * temp_1n / self.pamset["threstemp"])
                * self.pamset["dt"]
            )
        self.varrying["dtrm1n"] = (
            1.0
            - self.pamset["cpi"] * wcfac / self.dz[0]
            - self.pamset["beto"] * self.pamset["dt"] / (self.pamset["c1"] * self.dz[0])
        )
        self.varrying["dtmnl2"] = (
            wcfac * self.pamset["cpi"] / self.dz[self.pamset["lm"] - 1]
        )
        (
            self.varrying["acoeffn"],
            self.varrying["bcoeffn"],
            self.varrying["ccoeffn"],
        ) = self.coeff(wcfac, self.get_gam_and_fro_factor_ns(True))

        # Southern hemisphere:
        if self.pamset["threstemp"] == 0:  # pylint: disable=compare-to-zero
            wcfac = self.pamset["W"] / (SEC_DAY * DAY_YEAR) * self.pamset["dt"]
        else:
            wcfac = (
                self.pamset["W"]
                / (SEC_DAY * DAY_YEAR)
                * (1 - 0.3 * temp_1s / self.pamset["threstemp"])
                * self.pamset["dt"]
            )
        self.varrying["dtrm1s"] = (
            1.0
            - self.pamset["cpi"] * wcfac / self.dz[0]
            - self.pamset["fnso"]
            * self.pamset["beto"]
            * self.pamset["dt"]
            / (self.pamset["c1"] * self.dz[0])
        )
        self.varrying["dtmsl2"] = (
            wcfac * self.pamset["cpi"] / self.dz[self.pamset["lm"] - 1]
        )
        (
            self.varrying["acoeffs"],
            self.varrying["bcoeffs"],
            self.varrying["ccoeffs"],
        ) = self.coeff(wcfac, self.get_gam_and_fro_factor_ns(False))

    def setup_ebud(self):
        """
        Set up energy budget before run

        Find various paramaters at the start of the run,
        to get the energy budget ready to run
        """
        fnsa = 1.0  # Can it be something else?
        c1fac = self.pamset["dt"] / (self.pamset["c1"] * self.dz[0])

        blm = self.pamset["ebbeta"] / self.pamset["rlamdo"]
        self.gamn = (
            self.pamset["foan"] + self.pamset["rlamda"] / self.pamset["rlamdo"] + blm
        )
        self.gams = (
            self.pamset["foas"]
            + self.pamset["rlamda"] / self.pamset["rlamdo"]
            + fnsa * blm
        )

        # Northern hemisphere
        self.varrying["dtrm2n"] = (
            self.pamset["beto"]
            + self.pamset["foas"]
            * self.pamset["ebbeta"]
            / (self.gams * self.gamn - fnsa * blm**2)
        ) * c1fac
        self.varrying["dtrm3n"] = (
            self.gams / (self.gams * self.gamn - fnsa * blm**2) * c1fac
        )
        self.varrying["dtrm4n"] = blm / (self.gams * self.gamn - fnsa * blm**2) * c1fac

        # Southern hemisphere
        self.varrying["dtrm2s"] = (
            self.pamset["fnso"] * self.pamset["beto"]
            + self.pamset["foan"]
            * fnsa
            * self.pamset["ebbeta"]
            / (self.gams * self.gamn - fnsa * blm**2)
        ) * c1fac
        self.varrying["dtrm3s"] = (
            self.gamn / (self.gams * self.gamn - fnsa * blm**2) * c1fac
        )
        self.varrying["dtrm4s"] = (
            fnsa * blm / (self.gams * self.gamn - fnsa * blm**2) * c1fac
        )

        self.varrying["dtmnl3"] = (
            self.pamset["dt"]
            * self.pamset["beto"]
            / (self.pamset["c1"] * self.dz[self.pamset["lm"] - 1])
        )
        self.varrying["dtmnl1"] = 1.0 - self.varrying["dtmnl3"]
        self.varrying["dtmsl3"] = self.pamset["fnso"] * self.varrying["dtmnl3"]
        self.varrying["dtmsl1"] = 1.0 - self.varrying["dtmsl3"]
        self.setup_ebud2(0, 0)

    def energy_budget(
        self, forc_nh, forc_sh, fn_volc, fs_volc
    ):  # pylint: disable=too-many-locals
        """
        Do energy budget calculation for single year

        Parameters
        ----------
        forc_nh : float
               Northern hemispheric forcing
        forc_sh : float
               Southern hemispheric forcing
        fn_volc : float
               Northern hemispheric volcanic forcing
        fs_volc : float
               Northern hemispheric volcanic forcing

        Returns
        -------
        dict
            Containing all results including
            Global temperature change dtemp,
            Northern hemisphere temperature change dtempnh,
            Southern hemisphere temperature change dtempsh,
            Global air temperature change dtemp_air,
            Northern hemisphere air temperature change dtempnh_air,
            Southern hemisphere temperature change dtempsh_air,
            Global sea temperature change dtemp_sea,
            Northern hemisphere sea temperature change dtempnh_sea,
            Southern hemisphere sea temperature change dtempsh_sea,
            Northern hemisphere radiative imbalance RIBN,
            Southern hemisphere radiative imbalance RIBS,
            Global radiative imbalance RIB,
            Ocean heat content change down to 700 m OHC700,
            Ocean heat content change total OHCTOT
        """
        temp1n = 0.0
        temp1s = 0.0

        tempn = 0.0
        temps = 0.0
        tempn_air = 0.0
        temps_air = 0.0
        tempn_sea = 0.0
        temps_sea = 0.0
        lm = self.pamset["lm"]
        templ = np.zeros(lm)

        dtyear = 1.0 / self.pamset["ldtime"]
        dn = np.zeros(lm)
        ds = np.zeros(lm)
        for im in range(self.pamset["ldtime"]):
            volc_idx = im % len(fn_volc)
            if self.pamset["threstemp"] != 0:  # pylint: disable=compare-to-zero
                self.setup_ebud2(temp1n, temp1s)

            dqn = (
                (im + 1) * forc_nh * dtyear
                + (1 - (im + 1) * dtyear) * self.prev_values["fn"]
                + fn_volc[volc_idx]
            )
            dqs = (
                (im + 1) * forc_sh * dtyear
                + (1 - (im + 1) * dtyear) * self.prev_values["fs"]
                + fs_volc[volc_idx]
            )
            dn[0] = (
                self.varrying["dtrm1n"] * self.tn[0]
                + self.varrying["dtrm2n"] * self.ts[0]
                + self.varrying["dtrm3n"] * dqn
                + self.varrying["dtrm4n"] * dqs
            )
            ds[0] = (
                self.varrying["dtrm1s"] * self.ts[0]
                + self.varrying["dtrm2s"] * self.tn[0]
                + self.varrying["dtrm3s"] * dqs
                + self.varrying["dtrm4s"] * dqn
            )
            dn[1 : lm - 1] = self.tn[1 : lm - 1] + self.pamset["beto"] * self.pamset[
                "dt"
            ] / (self.pamset["c1"] * self.dz[1 : lm - 1]) * (
                self.ts[1 : lm - 1] - self.tn[1 : lm - 1]
            )
            ds[1 : lm - 1] = self.ts[1 : lm - 1] + self.pamset["fnso"] * self.pamset[
                "beto"
            ] * self.pamset["dt"] / (self.pamset["c1"] * self.dz[1 : lm - 1]) * (
                self.tn[1 : lm - 1] - self.ts[1 : lm - 1]
            )

            dn[lm - 1] = (
                self.varrying["dtmnl1"] * self.tn[lm - 1]
                + self.varrying["dtmnl2"] * self.tn[0]
                + self.varrying["dtmnl3"] * self.ts[lm - 1]
            )
            ds[lm - 1] = (
                self.varrying["dtmsl1"] * self.ts[lm - 1]
                + self.varrying["dtmsl2"] * self.ts[0]
                + self.varrying["dtmsl3"] * self.tn[lm - 1]
            )

            #
            # Where are these being initialised? Ok, I think
            self.tn = _band(
                self.varrying["acoeffn"],
                self.varrying["bcoeffn"],
                self.varrying["ccoeffn"],
                dn,
            )
            self.ts = _band(
                self.varrying["acoeffs"],
                self.varrying["bcoeffs"],
                self.varrying["ccoeffs"],
                ds,
            )
            # print("self.aceoffn: %s self.varrying["bcoeffn"]: %s self.ccoeffn %s"%(self.varrying["acoeffn"], self.varrying["bcoeffn"], self.varrying["ccoeffn"]))
            # print("self.aceoffs: %s self.bcoeffs: %s self.ccoeffs %s"%(self.varrying["acoeffs"], self.bcoeffs, self.ccoeffs))
            temp1n = self.tn[0]
            temp1s = self.ts[0]
            # print("temp1n: %.5e temp1s %.5e"%(temp1n, temp1s))
            templ = (
                templ + 0.5 * (self.tn + self.ts) / self.pamset["ldtime"]  # 12.0
            )  # used to be 12, now ldtime

            tempan = (
                dqn
                + self.pamset["foan"] * self.pamset["rlamdo"] * temp1n
                + self.pamset["ebbeta"]
                * (dqs + self.pamset["foas"] * self.pamset["rlamdo"] * temp1s)
                / self.pamset["fsx"]
            )
            tempan = tempan / (
                self.pamset["fnx"] - self.pamset["ebbeta"] ** 2 / self.pamset["fsx"]
            )
            tempas = (
                dqs
                + self.pamset["foas"] * self.pamset["rlamdo"] * temp1s
                + self.pamset["ebbeta"]
                * (dqn + self.pamset["foan"] * self.pamset["rlamdo"] * temp1n)
                / self.pamset["fnx"]
            )
            tempas = tempas / (
                self.pamset["fsx"] - self.pamset["ebbeta"] ** 2 / self.pamset["fnx"]
            )
            tmpn = self.pamset["foan"] * temp1n + (1.0 - self.pamset["foan"]) * tempan
            tmps = self.pamset["foas"] * temp1s + (1.0 - self.pamset["foas"]) * tempas

            # x1=1638.+float(years_since_start)+float(im-1)/12.

            tempn = tempn + tmpn / self.pamset["ldtime"]  # Previously 12.0
            temps = temps + tmps / self.pamset["ldtime"]  # Previously 12.0

            tempn_air = tempn_air + tempan / self.pamset["ldtime"]  # Previously 12.0
            temps_air = temps_air + tempas / self.pamset["ldtime"]  # Previously12.0

            tempn_sea = tempn_sea + temp1n / self.pamset["ldtime"]  # Previously 12.0
            temps_sea = temps_sea + temp1s / self.pamset["ldtime"]  # Previously 12

        dtemp = (tempn + temps) / 2.0  # Global temp chg)

        # Updating previous values for next year
        self.prev_values["fn"] = forc_nh
        self.prev_values["fs"] = forc_sh
        self.prev_values["dtemp"] = dtemp

        # Getting Ocean temperature:
        ocean_res = self.ocean_temperature()
        ribn = forc_nh + np.mean(fn_volc) - self.pamset["rlamda"] * tempn
        ribs = forc_sh + np.mean(fs_volc) - self.pamset["rlamda"] * temps
        # Returning results_dict
        return {
            "dtemp": dtemp,
            "dtempnh": tempn,
            "dtempsh": temps,
            "dtemp_air": (tempn_air + temps_air) / 2.0,
            "dtempnh_air": tempn_air,
            "dtempsh_air": temps_air,
            "dtemp_sea": (tempn_sea + temps_sea) / 2.0,
            "dtempnh_sea": tempn_sea,
            "dtempsh_sea": temps_sea,
            "RIBN": ribn,
            "RIBS": ribs,
            "RIB": (ribn + ribs) / 2.0,
            "OHC700": ocean_res["OHC700"],
            "OHCTOT": ocean_res["OHCTOT"],
        }

    def ocean_temperature(self):
        """
        Compute the ocean temperature total and at 700 m depth

        Compute the ocean temperature total and at 700 m depth
        to get ocean heat content for total and down to 700 m

        Returns
        -------
        dict
            Containing the ocean heat content to 700 m and total
            with keys OHC700 and OHCTOT
        """
        area_hemisphere = 2.55e14

        rho = 1030.0
        constant = 3.997e-19
        havtemp = (
            rho
            * constant
            * area_hemisphere
            * self.dz
            * (self.tn * self.pamset["foan"] + self.ts * self.pamset["foas"])
        )

        # Finding the max layer down to 700m
        max_layer = int(7 - self.dz[0] // 100.0)
        frac = (1 + self.dz[0] // 100.0) - self.dz[0] / 100.0

        return {
            "OHC700": np.sum(havtemp[:max_layer]) + frac * havtemp[max_layer],
            "OHCTOT": np.sum(havtemp[:]),
        }



import numpy as np

from ciceroscm.parallel._configdistro import _ConfigDistro


def test_config_distro_methods():
    config_full = _ConfigDistro(
        distro_array=np.array([[0, 1], [0, 1], [0, 1], [4, 6]]),
        setvalues={
            "qbmb": 0,
            "qo3": 0.5,
            "qdirso2": -0.36,
            "qindso2": -0.97,
            "threstemp": 7.0,
            "lm": 40,
            "ldtime": 12,
            "qh2o_ch4": 0.091915,
        },
    )

    assert np.array_equal(
        config_full.prior,
        np.array(
            [
                [0, 1],
                [0, 1],
                [0, 1],
                [4, 6],
                [0, 7],
                [2 / 3.71, 5 / 3.71],
                [25, 125],
                [0.1, 0.2],
                [-0.1, -0.06],
            ]
        ),
    )
    assert config_full.ordering == [
        "rlamdo",
        "akapa",
        "cpi",
        "W",
        "beto",
        "lambda",
        "mixed",
        "qbc",
        "qoc",
    ]

    num_latin = 100
    latin_samples = config_full.get_samples_from_distro_latin(num_latin)
    gaussian_samples = config_full.get_samples_from_distro_gaussian(50)
    assert latin_samples.shape == (100, len(config_full.ordering))
    assert gaussian_samples.shape == (50, len(config_full.ordering))
    for i, interval in enumerate(config_full.prior):
        masked = np.ma.masked_inside(latin_samples[:, i], interval[0], interval[1])
        assert np.ma.allequal(masked, [True] * num_latin)

    config_list = config_full.make_config_list(20, indexer_pre="test")
    assert len(config_list) == 20
    assert config_list[4]["Index"] == "test4"
    expected_emiconc = ["qbc", "qoc", "qbmb", "qdirso2", "qindso2", "qo3", "qh2o_ch4"]
    expected_udm = [
        "rlamdo",
        "akapa",
        "cpi",
        "W",
        "beto",
        "lambda",
        "mixed",
        "lm",
        "ldtime",
        "threstemp",
    ]
    assert all(pam in config_list[7]["pamset_udm"] for pam in expected_udm)
    print(config_list[13]["pamset_emiconc"])
    assert all(pam in config_list[13]["pamset_emiconc"] for pam in expected_emiconc)
